"""
    UvDiffEffects.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import os
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize, basinhopping
from molass_legacy.KekLib.SciPyCookbook import smooth
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import get_setting, get_beamline_name
from molass_legacy.Batch.StandardProcedure import StandardProcedure
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.Peaks.EghSupples import egh, compute_AB, d_egh
from MatrixData import simple_plot_3d
from molass_legacy.Elution.CurveUtils import simple_plot
from DataUtils import get_in_folder
from MplAnnotate import get_annotate_position
from LPM import get_corrected

WAVELENGTHS = [280, 400]    # ng at 420 for 20170209/OAGI_01

def get_uv_data_from_file(uv_file):
    from SerialDataUtils import load_uv_array
    from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
    folder, file = os.path.split(uv_file)
    D, wv, _ = load_uv_array(folder, file)
    print(file, D.shape)
    if D.shape[1] > 100:
        i1 = bisect_right(wv, WAVELENGTHS[0])
        y1 = D[i1,:]
        ecurve = ElutionCurve(y1)
    else:
        ecurve = None
    return D, wv, ecurve

debug_eval = False

def investigate_diff_effects(in_folder, use_spline=False, fh=None, show_fig=True, fig_file=None, logger=None, uv_file=None, ret_info=False):
    global debug_eval

    if uv_file is None:
        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)

        if get_setting("use_xray_conc"):
            if logger is not None:
                logger.info("skipping %s", in_folder)
            return

        if logger is not None:
            logger.info("doing %s", in_folder)
        pre_recog = PreliminaryRecognition(sd)
        D, _, wv, ecurve = sd.get_uv_data_separate_ly()
    else:
        # uv_file is used for those without X-ray data
        sd = None
        D, wv, ecurve = get_uv_data_from_file(uv_file)
        if ecurve is None:
            return

    i1, i2 = [bisect_right(wv, v) for v in WAVELENGTHS]

    x = ecurve.x
    y = ecurve.y

    y2 = D[i2,:]

    pn = ecurve.primary_peak_no
    emg_peaks = ecurve.get_emg_peaks()
    epk = emg_peaks[pn]
    # my = egh(x, *epk.opt_params)
    # my = epk.get_model_y(x)
    h, m, s, t = epk.get_params()
    my = egh(x, h, m, s, t)

    if use_spline:
        from molass_legacy.Trimming.PeakRegion import PeakRegion
        diff_spline = ecurve.d1
        def diff_model(x, scale):
            return scale * diff_spline(x)

        if sd is None:
            if True:
                ady = np.abs(diff_spline(x))
                ady = ady/np.max(ady)
                wh = np.where(ady > 0.05)[0]
                ilp, irp = wh[[0,-1]]
            else:
                ends_list = []
                for k, epk_ in enumerate(emg_peaks):
                    h_, m_, s_, t_ = epk_.get_params()
                    print("--------------------", [k], m_)
                    # my = egh(x, h_, m_, s_, t_)
                    A, B = compute_AB(np.log(0.03), s_, t_)
                    lp, rp = m_ - A, m_ + B
                    ends_list.append((lp, rp))

                ilp, irp = [int(p) for p in [ends_list[0][0], ends_list[-1][1]]]
        else:
            from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at

            xr_curve = sd.get_xray_curve()
            a_curve2 = make_secondary_e_curve_at(D, wv, ecurve, logger, wavelen=400)
            pr = PeakRegion(xr_curve, ecurve, a_curve2)
            # uv_peak_ends = pr.get_peak_ends([0.05, 0.95])[1]
            uv_peak_ends = pr.get_wider_ends()
            ilp, irp = uv_peak_ends
    else:
        def diff_model(x, scale):
            return d_egh(x, scale, m, s, t)

        A, B = compute_AB(np.log(0.05), s, t)
        lp, rp = m - A, m + B

        ilp, irp = [int(p) for p in [lp, rp]]

    pk_slice = slice(ilp, irp+1)
    print("pk_slice=", pk_slice)

    x_ = x[pk_slice]
    y2_ = y2[pk_slice]
    # sy2 = smooth(y2_, window_len=10)
    sy2 = smooth(y2_)

    def diff_func(x, scale, slope, intercept):
        return diff_model(x, scale) + slope * x + intercept

    def fit_func(p):
        return np.sum((diff_func(x_, *p) - sy2)**2)

    def evaluate(ret):
        global debug_eval
        scale = ret.x[0]
        scale_ratio = scale/h
        # dy = d_egh(x_, scale, m, s, t)
        dy = diff_model(x_, scale)
        height = np.max(dy) - np.min(dy)
        height_ratio = height/h
        print("scale_ratio=", scale_ratio)
        print("height_ratio=", height_ratio)
        rmsd = np.sqrt(np.average((diff_func(x_, *ret.x) - sy2)**2))
        print("rmsd=", rmsd)
        if debug_eval:
            slope, intercept = ret.x[1:3]
            with plt.Dp():
                fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,8))
                ax1 = axes[0,0]
                ax2 = axes[1,0]
                ax1.plot(x, y)
                ymin, ymax = ax1.get_ylim()
                ax1.set_ylim(ymin, ymax)
                for j in ilp, irp:
                    ax1.plot([j, j], [ymin, ymax], color="yellow")

                # ax1.plot(x, my, ":")
                ax2.plot(x, y2)
                ax2.plot(x_, sy2)
                ax2.plot(x_, dy + x_*slope + intercept, ":", color="cyan")

                ax3 = axes[0,1]
                ax4 = axes[1,1]

                for ax, scale_ in (ax3,1), (ax4, scale):
                    ax.plot(x, diff_model(x, scale_))
                    ax.plot(x_, diff_model(x_, scale_), color="cyan")

                fig.tight_layout()
                debug_eval = plt.show()
        return scale, height, rmsd

    results = []
    # method = None
    method = "Nelder-Mead"
    for init_scale in [-1, -0.02, -0.001, 0.001,  0.02, 1]:
        ret = minimize(fit_func, (init_scale, 0, 0), method=method)
        # ret = basinhopping(fit_func, (init_scale, 0, 0))
        scale, height, rmsd = evaluate(ret)
        results.append((abs(rmsd/height), ret))

    results = sorted(results, key=lambda p: p[0])
    print("results=", [r[0] for r in results])
    opt_ret = results[0][1]
    # debug_eval = True
    scale, height, rmsd = evaluate(opt_ret)

    in_folder_ = get_in_folder(in_folder)
    uv_device_no = get_setting("uv_device_no")

    rec = [in_folder_, uv_device_no] + [str(v)  for v in [h, scale, height, rmsd]]
    if fh is None:
        print(rec)
    else:
        fh.write(",".join(rec) + "\n")

    # h_ratio = height/h
    h_ratio = height/np.max(y)
    info = uv_file, ecurve, diff_func, x, my, ilp, irp, h, m, s, t, y2, x_, sy2, opt_ret, h_ratio

    if show_fig:

        with plt.Dp():
            filename = "" if uv_file is None else "/%s" % os.path.split(uv_file)[1]
            beamline = get_beamline_name(uv_device_no)
            fig = plt.figure(figsize=(18,6))
            fig.suptitle("Differential Effects Investigation in %s%s from %s" % (in_folder_, filename, beamline), fontsize=20)
            ax1 = fig.add_subplot(131, projection="3d")
            ax2 = fig.add_subplot(132)
            ax3 = fig.add_subplot(133)
            ax1.set_title("3D View", fontsize=16)
            ax2.set_title("Elution at λ=%d" % WAVELENGTHS[0], fontsize=16)
            ax3.set_title("Elution at λ=%d" % WAVELENGTHS[1], fontsize=16)

            simple_plot_3d(ax1, D, x=wv)
            uy = np.arange(len(y2))
            for i, c in [(i1, 'C1'), (i2, 'cyan')]:
                ux = np.ones(len(uy)) * wv[i]
                uz = D[i,:]
                ax1.plot(ux, uy, uz, color=c)

            draw_2d_figs(ax2, ax3, info)

            fig.tight_layout()
            fig.subplots_adjust(top=0.85)
            if fig_file is None:
                plt.show()
            else:
                from time import sleep
                fig.savefig(fig_file)
                plt.show(block=False)
                sleep(0.5)

    if ret_info:
        return info

def draw_2d_figs(ax2, ax3, info):
    uv_file, ecurve, diff_func, x, my, ilp, irp, h, m, s, t, y2, x_, sy2, opt_ret, h_ratio = info

    if True:
        ax2.plot(ecurve.x, ecurve.y)
        ymin, ymax = ax2.get_ylim()
        ax2.set_ylim(ymin, ymax)
        for j in ilp, irp:
            ax2.plot([j, j], [ymin, ymax], color="yellow")
    else:
        simple_plot(ax2, ecurve, legend=False)
        ax2.plot(x, my, ":", label="egh model", lw=3)
        for p in [ilp, irp]:
            ax2.plot(p, egh(np.array([p]), h, m, s, t), "o", color="yellow")
        ax2.legend()

    ax3.plot(x, y2, label="data")
    ax3.plot(x_, sy2, label="data")
    ax3.plot(x_, diff_func(x_, *opt_ret.x), ":", label="fit diff", lw=3, color="cyan")
    ymin, ymax = ax3.get_ylim()
    ax3.set_ylim(ymin, ymax)
    for j in ilp, irp:
        ax3.plot([j, j], [ymin, ymax], color="yellow")
    ax3.legend()

    tx, ty, r = get_annotate_position(ax3, debug=False)
    ax3.text(tx, ty, "ratio=%.2g" % h_ratio, ha="center", va="center", alpha=0.5, fontsize=20)



"""
    DraggablePoints.py

    borrowed and adapted from
        Matplotlib drag overlapping points interactively
        http://stackoverflow.com/questions/21654008/matplotlib-drag-overlapping-points-interactively
"""
import matplotlib.pyplot as plt
import matplotlib.patches as patches

class DraggablePoints(object):
    def __init__( self, artists, tolerance=5, y_only=False,
                    select_cb=None, update_cb=None, select_fc=None ):
        for artist in artists:
            artist.set_picker(tolerance)
        self.artists = artists
        self.axes   = artists[0].axes
        self.artists_dict = {}
        for i, artist in enumerate( artists ):
            self.artists_dict[artist] = [ i, artist.get_facecolor() ]
        self.currently_dragging = False
        self.has_moved = False
        self.current_artist = None
        self.offset = (0, 0)
        self.y_only = y_only
        self.select_cb = select_cb
        self.update_cb = update_cb
        self.select_fc = select_fc

        for canvas in set(artist.figure.canvas for artist in self.artists):
            canvas.mpl_connect('button_press_event', self.on_press)
            canvas.mpl_connect('button_release_event', self.on_release)
            canvas.mpl_connect('pick_event', self.on_pick)
            canvas.mpl_connect('motion_notify_event', self.on_motion)

    def on_press(self, event):
        if event.inaxes != self.axes:
            return

        self.currently_dragging = True
        self.has_moved = False

    def on_release(self, event):
        self.currently_dragging = False
        self.current_artist = None
        if self.update_cb is not None and self.has_moved:
            self.update_cb( self )

    def on_pick(self, event):
        if self.current_artist is None:
            self.current_artist = event.artist
            # self.current_artist.set( alpha=1 )
            if self.select_cb is not None:
                self.select_cb( self.artists_dict[event.artist][0] )
            x0, y0 = event.artist.center
            x1, y1 = event.mouseevent.xdata, event.mouseevent.ydata
            self.offset = (x0 - x1), (y0 - y1)
            if self.select_fc is not None:
                for a in self.artists:
                    if a == self.current_artist:
                        a.set( fc=self.select_fc )
                    else:
                        a.set( fc=self.artists_dict[a][1] )
                self.current_artist.figure.canvas.draw()

    def on_motion(self, event):
        if not self.currently_dragging:
            return
        if self.current_artist is None:
            return
        dx, dy = self.offset
        if self.y_only:
            x_, y_ = self.current_artist.center
            self.current_artist.center = x_, event.ydata + dy
        else:
            self.current_artist.center = event.xdata + dx, event.ydata + dy
        self.current_artist.figure.canvas.draw()
        self.has_moved = True

    def unselect( self ):
        if self.select_fc is not None:
            for a in self.artists:
                    a.set( fc=self.artists_dict[a][1] )

    def select( self, ind ):
        for i, a in enumerate( self.artists ):
            if i == ind:
                fc_ = self.select_fc
            else:
                fc_ = self.artists_dict[a][1]
            a.set( fc=fc_ )

if __name__ == '__main__':
    fig, ax = plt.subplots()
    ax.set(xlim=[-1, 2], ylim=[-1, 2])
    # ax.set(xlim=[ 0, 400 ], ylim=[-1, 2])

    circles = [patches.Circle((0.32, 0.3), 0.2, fc='r', alpha=0.5),
               patches.Circle((0.3, 0.3), 0.2, fc='b', alpha=0.5) ]
    for circ in circles:
        ax.add_patch(circ)

    def select_cb( ind ):
        print( 'select_cb: ind=', ind )

    def update_cb( obj ):
        print( 'update_cb' )

    dr = DraggablePoints( circles, select_cb=select_cb, update_cb=update_cb, select_fc='yellow' )
    plt.show()


"""
    HistoryDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
import matplotlib.pyplot as plt

this_dir = os.path.dirname(os.path.abspath( __file__ ))
sys.path.append(this_dir + '/..')
import molass_legacy.KekLib
from molass_legacy.Models.ElutionCurveModels import egh, emg
from SecTheory.Edm import guess_single_edm, edm_func

def demo():
    x = np.arange(300)

    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(15,3))
    for ax in ax1, ax2, ax3:
        ax.set_axis_off()

    ax1.plot(x, egh(x, 1, 150, 20, 0), label="Gaussian")
    ax2.plot(x, egh(x, 1, 120, 20, 0), ":", label="Gaussian")
    h = emg(np.array([120]), 1, 105, 18, 20)
    ax2.plot(x, emg(x, 1/h[0], 105, 18, 20), label="EMG")
    ax2.plot(x, egh(x, 1, 120, 20, 20), label="EGH")

    egh_y = egh(x, 1, 120, 20, 20)
    edm = guess_single_edm(x, egh_y)
    params = edm.get_comp_params()
    u = 0.5
    edm_y = edm_func(x, u, *params)

    ax3.plot(x, egh_y, ":", label="EGH", color="C2")
    ax3.plot(x, edm_y, label="EDM", color="C3")

    a, b, e, Dz, cinj = params
    edm_y2 = edm_func(x, u, a/3.5, -3*b, e, Dz, cinj)
    k = np.argmax(edm_y2)
    ax3.plot(x, edm_y2/edm_y2[k], label="fronting EDM", color="C4")

    for ax in ax1, ax2, ax3:
        ax.legend()

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    demo()



# coding: utf-8
"""
    SaferSpinbox.py

    Copyright (c) 2018-2019, Masatsuyo Takahashi, KEK-PF
"""
from time import time
import tkinter as Tk

class SaferSpinbox(Tk.Spinbox):
    def __init__(self, parent, **kwargs):
        self.inverval = kwargs.pop("inverval", 1)
        command = kwargs.pop("command", None)
        if command is None:
            command = self.command
        Tk.Spinbox.__init__(self, parent, command=command, **kwargs)
        self.user_tracer = None
        self.var = kwargs.get("textvariable", None)
        self.from_ = kwargs.get("from_", 1)
        self.to = kwargs.get("to", 9)
        self.change_history = []

    def config(self, **kwargs):
        # todo: better overriding
        self.var = kwargs.get("textvariable", self.var)
        self.from_ = kwargs.get("from_", self.from_)
        self.to = kwargs.get("to", self.to)
        Tk.Spinbox.config(self, **kwargs)

    def command(self):
        if self.user_tracer is not None:
            # always call user_tracer on arrow button click
            self.user_tracer(*self.last_args)
            self.change_history = []

    def set_tracer(self, tracer):
        self.user_tracer = tracer

    def tracer(self, *args):
        try:
            value = self.var.get()
        except:
            # from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            # etb = ExceptionTracebacker()
            # print(etb.last_lines())
            return

        self.last_value = value
        self.last_args = args
        self.change_history.append(time())
        if self.user_tracer is not None:
            self.after(1100, self.user_tracer_caller)

    def user_tracer_caller(self):
        if len(self.change_history) == 0:
            return

        interval = time() - self.change_history[-1]
        if interval > self.inverval:
            if self.validate():
                self.user_tracer(*self.last_args)
                self.change_history = []
        else:
            self.error_notified = False
            self.after(100, self.user_tracer_caller)

    def validate(self):
        try:
            value = self.var.get()
            if self.from_  <= value and value <= self.to:
                return True

            if not self.error_notified:
                import molass_legacy.KekLib.CustomMessageBox as MessageBox
                MessageBox.showerror( "Value Error",
                    "Values for this spinbox must be within %d and %d." % (self.from_, self.to),
                    parent=self)
                self.error_notified = True
        except:
            pass

        return False



# coding: utf-8
"""
    GuinierViewer.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from CanvasFrame            import CanvasFrame
from molass_legacy.KekLib.OurMatplotlib          import get_default_colors
from SimpleGuinier          import SimpleGuinier

class GuinierFrame( Tk.Frame ):
    def __init__( self, parent ):
        Tk.Frame.__init__( self, parent)

        figsize = ( 15, 5 ) if is_low_resolution() else ( 18, 6 )
        self.canvas_frame = canvas_frame = CanvasFrame( self, figsize=figsize )
        canvas_frame.pack()

        fig = canvas_frame.fig
        ax1  = fig.add_subplot( 131 )
        ax2  = fig.add_subplot( 132 )
        ax3  = fig.add_subplot( 133 )
        fig.tight_layout()
        self.axes = [ax1, ax2, ax3]

    def draw( self, xx, ly1, ly2, zoom=False, colors=None, gcolors=None, guinier_list=None ):
        for ax in self.axes:
            ax.cla()

        ax1, ax2, ax3 = self.axes

        if colors is None:
            color1, color2 = None, None
        else:
            color1, color2 = colors

        if gcolors is None:
            gcolor1, gcolor2 = 'cyan', 'red'
        else:
            gcolor1, gcolor2 = gcolors

        ax1.plot( xx, ly1, 'o', markersize=3, color=color1 )
        ax2.plot( xx, ly2, 'o', markersize=3, color=color2 )

        ax3.plot( xx, ly1, 'o', markersize=3, color=color1 )
        ax3.plot( xx, ly2, 'o', markersize=3, color=color2 )

        if zoom:
            xmin1, xmax1 = ax1.get_xlim()
            ymin1, ymax1 = ax1.get_ylim()
            xmin_   = xmin1 * 0.98 + xmax1 * 0.02
            xmax_   = xmin1 * 0.80 + xmax1 * 0.2
            # ymin_   = ymin1 * 0.3  + ymax1 * 0.7
            # ymax_   = ymin1 * 0.05 + ymax1 * 0.95
            ymin_, ymax_ = -9.0, -3.0
            for ax in self.axes:
                ax.set_xlim( xmin_, xmax_ )
                ax.set_ylim( ymin_, ymax_ )

        if guinier_list is not None:
            guinier1 = guinier_list[0]
            x_  = guinier1.guinier_x
            y_  = guinier1.guinier_y
            for ax in [ax1, ax3]:
                ax.plot( x_, y_, color=gcolor1, marker='o', markersize=8 )

            guinier2 = guinier_list[1]
            x_  = guinier2.guinier_x
            y_  = guinier2.guinier_y
            for ax in [ax2, ax3]:
                ax.plot( x_, y_, color=gcolor2, marker='o', markersize=8 )

            tx  = xmin_*0.6 + xmax_*0.4
            ty  = ymin_*0.5 + ymax_*0.5
            ax1.text( tx, ty, "Rg=%.3g" % guinier1.Rg, alpha=0.2, fontsize=30 )
            ax2.text( tx, ty, "Rg=%.3g" % guinier2.Rg, alpha=0.2, fontsize=30 )
            ax3.text( tx, ty, "Rg diff=%.3g" % ( guinier2.Rg - guinier1.Rg ), alpha=0.2, fontsize=30 )

        self.canvas_frame.show()

class GuinierViewer( Tk.Toplevel ):
    def __init__( self, parent, qvector, data1, data2, data3, j ):
        self.parent = parent
        Tk.Toplevel.__init__( self, parent )
        self.title( "GuinierViewer" )

        self.data1  = data1
        self.data2  = data2
        self.data3  = data3

        vsize   = len(qvector)//2
        slice_  = slice(0,vsize)
        self.slice_ = slice_

        self.xx  = qvector[slice_]**2

        plot_frame = Tk.Frame( self )
        plot_frame.pack( side=Tk.LEFT )
        panel_frame = Tk.Frame( self )
        panel_frame.pack( side=Tk.LEFT, padx=20, pady=20 )

        self.gframe1 = GuinierFrame( plot_frame )
        self.gframe1.pack()
        self.gframe2 = GuinierFrame( plot_frame )
        self.gframe2.pack()

        jsize   = data1.shape[0]
        self.j  = Tk.IntVar()
        self.j.set( j )
        sinbox = Tk.Spinbox( panel_frame, textvariable=self.j,
                                            from_=0, to=jsize-1, increment=1, 
                                            justify=Tk.CENTER, width=6, state=Tk.NORMAL )
        sinbox.pack()

        redraw_button = Tk.Button( panel_frame, text="Redraw", command=self.draw )
        redraw_button.pack()

        self.draw()

        X   = parent.winfo_rootx()
        Y   = parent.winfo_rooty()
        print( 'X,Y=', X, Y )

        self.geometry("+%d+%d" % (X+50, Y+50))
        self.update()

    def get_guinier_data( self, j ):

        y1  = self.data1[j,self.slice_,1]
        y2  = self.data2[j,self.slice_,1]
        y3  = self.data3[j,self.slice_,1]

        ly1 = np.log( y1 )
        ly2 = np.log( y2 )
        ly3 = np.log( y3 )

        guinier_list = []
        for data in [ self.data1, self.data2, self.data3 ]:
            data_ = data[j,:,:]
            sg = SimpleGuinier( data_ )
            guinier_list.append( sg )

        return ly1, ly2, ly3, guinier_list

    def draw( self ):
        try:
            j   = self.j.get()
        except:
            return

        colors  = get_default_colors()
        gcolors = [ 'cyan', 'yellow', 'red' ]

        ly1, ly2, ly3, guinier_list = self.get_guinier_data( j )

        self.gframe1.draw( self.xx, ly1, ly2, zoom=True, colors=colors[0:2], gcolors=gcolors[0:2], guinier_list=guinier_list[0:2] )
        self.gframe2.draw( self.xx, ly2, ly3, zoom=True, colors=colors[1:3], gcolors=gcolors[1:3], guinier_list=guinier_list[1:3] )
        


# coding: utf-8
"""
    UnfoldingModel.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from lmfit import Parameters, minimize

RT = 1

class UnfoldingModel:
    def __init__(self, params=None):
        if params is None:
            params = Parameters()
            params.add('G', value=4)
            params.add('m', value=1)
            params.add('af', value=0.5)
            params.add('bf', value=-0.01)
            params.add('au', value=0.3)
            params.add('bu', value=-0.02)
        self.parems = params

    def compute_Pf(self, params, x):
        if params is None:
            params = self.parems
        G = params['G']
        m =  params['m']
        Pf = 1/(1 + np.exp(-(G-m*x)/RT))
        return Pf

    def compute_yf(self, params, Pf, x):
        if params is None:
            params = self.parems

        af = params['af']
        bf = params['bf']
        yf = (af + bf*x)*Pf
        return yf

    def compute_yu(self, params, Pf, x):
        if params is None:
            params = self.parems

        au = params['au']
        bu = params['bu']
        yu = (au + bu*x)*(1-Pf)
        return yu

    def fit(self, x, y):
        params = Parameters()
        params.add('G', value=2)
        params.add('m', value=2)
        params.add('af', value=0)
        params.add('bf', value=0)
        params.add('au', value=0)
        params.add('bu', value=0)

        def obj_func(p):
            Pf = self.compute_Pf(p, x)
            yf = self.compute_yf(p, Pf, x)
            yu = self.compute_yu(p, Pf, x)
            return yf+yu - y

        res = minimize(obj_func, params, args=())
        return res



# coding: utf-8
"""

    Canvas3D.py

        Canvas for 3D-plot with ayx-span selectors

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
import copy
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.gridspec as gridspec
from matplotlib.collections import PolyCollection
from matplotlib.widgets import Button, SpanSelector
from matplotlib import colors as mcolors

SPAN_STAYS  = False

class OurSpanSelector(SpanSelector):
    def __init__(*args, **kwargs):
        self = args[0]
        self.onpress_callback = kwargs.pop('onpress_callback', None)
        self.onrelease_callback = kwargs.pop('onrelease_callback', None)
        self.onhover_callback = kwargs.pop('onhover_callback', None)
        SpanSelector.__init__(*args, **kwargs)

    def _press(self, event):
        """on button press event"""
        if self.onpress_callback is not None:
            self.onpress_callback(event)
        return SpanSelector._press(self, event)

    def _release(self, event):
        """on button release event"""
        if self.onrelease_callback is not None:
            self.onrelease_callback(event)
        return SpanSelector._release(self, event)

    def onmove(self, event):
        """ overriding _SelectorWidget.onmove """
        if self.onhover_callback is not None:
            if not self.ignore(event):
                if self.pressv is None:
                    self.onhover_callback(event)
        return super(SpanSelector, self).onmove(event)

axis_names = [ "X", "Y", "Z" ]

# https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
default_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
dc = default_colors[0]
facecolors = [ mcolors.to_rgba(arg, alpha=0.1) for arg in [dc, dc] ]

class Canvas3D:
    def __init__(self, fig, grid_rows, grid_cols, hspace=0.3):
        self.fig = fig
        gs = gridspec.GridSpec( grid_rows, grid_cols )
        self.ax = fig.add_subplot(gs[0:grid_rows-3,:], projection='3d')
        axx = fig.add_subplot( gs[grid_rows-3,1:-1] )
        axy = fig.add_subplot( gs[grid_rows-2,1:-1] )
        axz = fig.add_subplot( gs[grid_rows-1,1:-1] )
        self.axes = [axx, axy, axz]

        fig.subplots_adjust( top=0.95, bottom=0.05, left=0.05, right=0.95, hspace=hspace )

        for k, ax_ in enumerate(self.axes):
            ax_.tick_params(
                which='both',
                bottom=False,
                top=False,
                left=False,
                labelbottom=False,
                labelleft=False)
            ax_.set_xlim(0, 1)
            ax_.text( -0.03, 0, axis_names[k] + " range selector", ha='right' )

        resetax = plt.axes([0.82, 0.18, 0.12, 0.04])
        self.v_reset_btn = Button(resetax, 'view reset', hovercolor='0.975')
        self.v_reset_btn.on_clicked(self.view_reset)

        resetax = plt.axes([0.82, 0.05, 0.12, 0.04])
        self.r_reset_btn = Button(resetax, 'range reset', hovercolor='0.975')
        self.r_reset_btn.on_clicked(self.range_reset)

        props = dict(alpha=0.5, facecolor=default_colors[0])
        minspan = 0.05

        spanx = OurSpanSelector(axx, self.on_select_x,
                    'horizontal', useblit=True, span_stays=SPAN_STAYS,
                    onpress_callback=self.on_press_or_hover_x,
                    onmove_callback=self.on_move_x,
                    onrelease_callback=self.on_release_x,
                    onhover_callback=self.on_press_or_hover_x,
                    props=props,
                    minspan=minspan)

        spany = OurSpanSelector(axy, self.on_select_y,
                    'horizontal', useblit=True, span_stays=SPAN_STAYS,
                    onpress_callback=self.on_press_or_hover_y,
                    onmove_callback=self.on_move_y,
                    onrelease_callback=self.on_release_y,
                    onhover_callback=self.on_press_or_hover_y,
                    props=props,
                    minspan=minspan)

        spanz = OurSpanSelector(axz, self.on_select_z,
                    'horizontal', useblit=True, span_stays=SPAN_STAYS,
                    onpress_callback=self.on_press_or_hover_z,
                    onmove_callback=self.on_move_z,
                    onrelease_callback=self.on_release_z,
                    onhover_callback=self.on_press_or_hover_z,
                    props=props,
                    minspan=minspan)

        fig.canvas.mpl_connect('axes_leave_event', self.on_leave_selector)
        fig.canvas.mpl_connect('motion_notify_event', self.on_leave_selector)
        fig.canvas.mpl_connect('figure_leave_event', self.on_leave_selector)

        self.selectors = [ spanx, spany, spanz ]
        self.init_limits = None
        self.poly_x = None
        self.poly_y = None
        self.poly_z = None

    def get_axis(self):
        return self.ax

    def set_init_lims(self):
        if self.init_limits is None:
            ax = self.ax
            self.init_limits = copy.deepcopy( [ ax.get_xlim(), ax.get_ylim(), ax.get_zlim() ] )
            print( 'init_limits=', self.init_limits )

    def view_reset(self, event):
        self.ax.view_init()
        self.fig.canvas.draw()

    def range_reset(self, event):
        # print( 'range_reset' )
        for sel in self.selectors:
            if SPAN_STAYS:
                sel.stay_rect.set_visible(False)
            sel.canvas.draw()

        if self.init_limits is not None:
            # print('reset limits')
            ax = self.ax
            ax.set_xlim( self.init_limits[0] )
            ax.set_ylim( self.init_limits[1] )
            ax.set_zlim( self.init_limits[2] )
            self.fig.canvas.draw()

    def on_select_x(self, min_, max_):
        # print( 'on_select_x', min_, max_ )
        xmin, xmax = self.ax.get_xlim()
        xmin_ = xmin * (1 - min_) + xmax * min_
        xmax_ = xmin * (1 - max_) + xmax * max_
        self.ax.set_xlim(xmin_, xmax_)
        self.fig.canvas.draw()

    def on_select_y(self, min_, max_):
        # print( 'on_select_y', min_, max_ )
        ymin, ymax = self.ax.get_ylim()
        ymin_ = ymin * (1 - min_) + ymax * min_
        ymax_ = ymin * (1 - max_) + ymax * max_
        self.ax.set_ylim(ymin_, ymax_)
        self.fig.canvas.draw()

    def on_select_z(self, min_, max_):
        # print( 'on_select_z', min_, max_ )
        zmin, zmax = self.ax.get_zlim()
        zmin_ = zmin * (1 - min_) + zmax * min_
        zmax_ = zmin * (1 - max_) + zmax * max_
        self.ax.set_zlim(zmin_, zmax_)
        self.fig.canvas.draw()

    def compute_zs(self, u, w, xminmax):
        self.set_init_lims()

        xmin, xmax = xminmax
        px = xmin * (1 - u) + xmax * u
        if w is None:
            w = u + 0.01
        qx = xmin * (1 - w) + xmax * w
        xs = [px, qx]
        return xs

    def compute_verts(self, yminmax, zminmax):
        ymin, ymax = yminmax
        zmin, zmax = zminmax
        y = [ymin, ymax, ymax, ymin]
        z = [zmin, zmin, zmax, zmax]
        yz = list(zip(y, z))
        verts = [ yz, yz ]
        return verts

    def on_press_or_hover_x(self, event):
        # print('on_press_x: ', event.xdata, event.ydata)
        ax = self.ax
        self.clear_poly_any(except_poly=0)
        if self.poly_x is not None:
            self.poly_x.remove()
        self.verts_x = self.compute_verts(ax.get_ylim(), ax.get_zlim())
        self.poly_x = PolyCollection(self.verts_x, facecolors=facecolors)
        xs = self.compute_zs(event.xdata, None, ax.get_xlim())
        ax.add_collection3d(self.poly_x, zs=xs, zdir='x')
        self.fig.canvas.draw()

    def on_press_or_hover_y(self, event):
        # print('on_press_y: ', event.xdata, event.ydata)
        ax = self.ax
        self.clear_poly_any(except_poly=1)
        if self.poly_y is not None:
            self.poly_y.remove()
        self.verts_y = self.compute_verts(ax.get_xlim(), ax.get_zlim())
        self.poly_y = PolyCollection(self.verts_y, facecolors=facecolors)
        ys = self.compute_zs(event.xdata, None, ax.get_ylim())
        ax.add_collection3d(self.poly_y, zs=ys, zdir='y')
        self.fig.canvas.draw()

    def on_press_or_hover_z(self, event):
        # print('on_press_z: ', event.xdata, event.ydata)
        ax = self.ax
        self.clear_poly_any(except_poly=2)
        if self.poly_z is not None:
            self.poly_z.remove()
        self.verts_z = self.compute_verts(ax.get_xlim(), ax.get_ylim())
        self.poly_z = PolyCollection(self.verts_z, facecolors=facecolors)
        zs = self.compute_zs(event.xdata, None, ax.get_zlim())
        ax.add_collection3d(self.poly_z, zs=zs, zdir='z')
        self.fig.canvas.draw()

    def on_move_x(self, min_, max_):
        ax = self.ax
        xs = self.compute_zs(min_, max_, ax.get_xlim())
        if self.poly_x is not None:
            self.poly_x.remove()
        self.verts_x = self.compute_verts(ax.get_ylim(), ax.get_zlim())
        self.poly_x = PolyCollection(self.verts_x, facecolors=facecolors)
        ax.add_collection3d(self.poly_x, zs=xs, zdir='x')
        self.fig.canvas.draw()

    def on_move_y(self, min_, max_):
        ax = self.ax
        ys = self.compute_zs(min_, max_, ax.get_ylim())
        if self.poly_y is not None:
            self.poly_y.remove()
        self.verts_y = self.compute_verts(ax.get_xlim(), ax.get_zlim())
        self.poly_y = PolyCollection(self.verts_y, facecolors=facecolors)
        ax.add_collection3d(self.poly_y, zs=ys, zdir='y')
        self.fig.canvas.draw()

    def on_move_z(self, min_, max_):
        ax = self.ax
        zs = self.compute_zs(min_, max_, ax.get_zlim())
        if self.poly_z is not None:
            self.poly_z.remove()
        self.verts_z = self.compute_verts(ax.get_xlim(), ax.get_ylim())
        self.poly_z = PolyCollection(self.verts_z, facecolors=facecolors)
        ax.add_collection3d(self.poly_z, zs=zs, zdir='z')
        self.fig.canvas.draw()

    def on_release_x(self, event):
        # print('on_release_x')
        if self.poly_x is not None:
            self.poly_x.remove()
        self.poly_x = None

    def on_release_y(self, event):
        # print('on_release_y')
        if self.poly_y is not None:
            self.poly_y.remove()
        self.poly_y = None

    def on_release_z(self, event):
        # print('on_release_z')
        if self.poly_z is not None:
            self.poly_z.remove()
        self.poly_z = None

    def clear_poly_any(self, except_poly=-1):
        removed = False
        if self.poly_x is not None and except_poly != 0:
            self.poly_x.remove()
            self.poly_x = None
            removed = True
        if self.poly_y is not None and except_poly != 1:
            self.poly_y.remove()
            self.poly_y = None
            removed = True
        if self.poly_z is not None and except_poly != 2:
            self.poly_z.remove()
            self.poly_z = None
            removed = True
        if removed:
            self.fig.canvas.draw()

    def on_leave_selector(self, event):
        if event.xdata is None or event.inaxes == self.ax:
            self.clear_poly_any()



"""
    ExtrapolationSolver.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import os
import copy
import time
import numpy as np
import logging
from scipy import optimize
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data, get_denoised_error
from .KnownPeakPenalty import compute_known_error_impl, compute_known_error_grad_impl
from molass_legacy.Conc.ConcDepend import compute_distinct_cd
from molass_legacy.LRF.LrfInfo import LrfInfo
import molass_legacy.KekLib.DebugPlot as plt

# SA_DEBUGGING = os.environ.get('SA_DEBUGGING', None)
SA_DEBUGGING = False

DEBUG   = False
DEBUG_WEIGHT_MATRIX = False
DEBUG_LRF_RANK = False

USE_TILTED_MODEL        = True
USE_PENALTY_PEAK_ERROR  = False
USE_LOGARITHMIC_RATIO   = False
LOGARITHMIC_RATIO       = 0.5
VERY_SMALL_VALUE        = 1e-5
BASE_ERROR_NORM_DEGREE  = 2
assert BASE_ERROR_NORM_DEGREE == 2
MAX_ITERATION_FMIN_CG   = 10000
USE_UNIFORMLY_SCALED_UV = True
ERROR_PROPAGATION_DENOISE   = False
DISABLE_IGNORE_BQ_LIST  = True
USE_DICTINCT_CD         = True

class KnownInfo:
    def __init__(self, data):
        self.data = data

    def __repr__(self):
        return str(self.data.shape)

class ExtrapolationSolver:
    def __init__( self, pdata, popts):
        self.rank_control = get_setting('rank_control')
        self.concentration_datatype = get_setting('concentration_datatype')

        self.judge_holder = pdata.judge_holder
        self.mapper = pdata.mapper          # self.mapper will be used in ElutionMatrix

        if pdata.is_for_sec:
            ecurve = self.mapper.x_curve
            self.sd = sd = pdata.sd
            self.qv = sd.qvector
            self.Rg = sd.pre_recog.get_rg()
            data = sd.intensity_array[:,:,1].T
            error = sd.intensity_array[:,:,2].T
            mc_vector = sd.mc_vector
            conc = None
            self.sd_id_info = sd.get_id_info()
            self.cd_slice = sd.get_cd_slice()
            self.xray_scale = sd.get_xray_scale()
        else:
            xdata = pdata.xdata
            self.sd = None
            self.qv = xdata.vector
            self.Rg = None
            ecurve = xdata.e_curve
            data = xdata.data
            error = xdata.error.data
            mc_vector = None
            conc = pdata.decomp_info.conc
            self.sd_id_info = None
            self.rank_control = 0   # not yet supported
            self.cd_slice = None
            self.xray_scale = None

        print( 'data.shape=', data.shape )

        if SA_DEBUGGING:
            from .ExtrapolationDebugger import ExtrapolationDebugger
            self.debugger = ExtrapolationDebugger()
        else:
            self.debugger = None

        self.logger = logging.getLogger( __name__ )
        self.ecurve = ecurve
        self.data = data
        self.error = error
        self.mc_vector = mc_vector  # already scaled with conc_factor
        self.cnv_ranges = pdata.cnv_ranges
        self.num_ranges = pdata.num_ranges
        self.conc_factor = pdata.conc_factor
        rg_str = "None" if self.Rg is None else "%.3g" % self.Rg
        self.logger.info('lrf solver has been constructed with conc_factor %g, Rg=%s.', self.conc_factor, rg_str)
        self.DEFAULT_WEIGHTS = get_setting('penalty_weighting')

        self.popts = popts
        self.aq_smoothness = popts.aq_smoothness
        self.aq_positivity = popts.aq_positivity
        self.use_elution_models = popts.use_elution_models
        self.conc = conc
        self.add_conc_const = get_dev_setting('add_conc_const')
        self.conc_dependence = get_setting('conc_dependence')
        self.doing_sec = self.conc is None
        if self.doing_sec:
            self.conc_depend = self.conc_dependence
        else:
            denat_dependent = get_setting('denat_dependent')
            self.conc_depend = 2 if denat_dependent else 1

        self.j0 = pdata.sd.xr_j0

        self.synthesized_lrf = get_setting('synthesized_lrf')
        self.logger.info('conc_depend has been set to %d' % self.conc_depend)

    def extrapolate_wiser( self, start, stop, peakset_info,
                            ignore_bq=0,
                            penalty_weights=None,
                            stop_event=None,
                            animation=False,
                            demo_info=None,
                            surplus_rank=0,
                            suppress_rank_control=False,
                            lrf_rank=None,
                            conc_depend=None,
                            save_cmatrix=False,
                            debug=False,
                            ):
        if debug:
            from importlib import reload
            import molass_legacy.LRF.ConcMatrix
            reload(molass_legacy.LRF.ConcMatrix)       
        from molass_legacy.LRF.ConcMatrix import ConcMatrix

        paired_ranges = self.cnv_ranges

        if penalty_weights is None:
            penalty_weights = copy.deepcopy(self.DEFAULT_WEIGHTS)

        self.penalty_weights = np.array(penalty_weights)
        self.stop_event = stop_event
        self.animation = animation
        self.demo_info = demo_info
        if animation:
            self.anim_data1 = []
            self.anim_data2 = []
            self.anim_counter = 0

        pno, nth, peakset, known_peak_info = peakset_info
        """
        NOTE:
            known_peak_info has been later added to [pno, nth, peakset],
            which has been appended to PeakSetSelector.required_peakset_infos.
            this differece in peakset_info, i.e., the difference between
            [pno, nth, peakset] and [pno, nth, peakset, known_peak_info],
            can be confusing.
        """
        paired_ranges_ = [paired_ranges[i] for i in peakset]
        # print('self.use_elution_models=', self.use_elution_models)
        # print('paired_ranges_=', paired_ranges_, 'ignore_bq=', ignore_bq)

        x = self.ecurve.x[start:stop]
        if self.mc_vector is None:
            mc_vector = None
        else:
            mc_vector = self.mc_vector[start:stop]

        if suppress_rank_control:
            rank_control = 0
        else:
            rank_control = self.rank_control

        if conc_depend is None:
            if rank_control:
                if lrf_rank is None:
                    if USE_DICTINCT_CD:
                        cmatrix = ConcMatrix(x, self.conc, conc_depend=1,
                                                paired_ranges=paired_ranges_, mc_vector=mc_vector,
                                                conc_factor=self.conc_factor,
                                                ecurve=self.ecurve, j0=self.j0,
                                                )
                        M = self.data[:,start:stop]
                        conc_depend = compute_distinct_cd(M, cmatrix.data, self.cd_slice, self.xray_scale, self.logger)
                    else:
                        conc_depend = self.judge_holder.get_cd_degree_from_range(start, stop)
                else:
                    if lrf_rank == 1:
                        conc_depend = 1
                    else:
                        conc_depend = 2
            else:
                conc_depend = self.conc_depend

        self.ignore_bq = 1 if conc_depend < 2 else 0

        cmatrix = ConcMatrix(x, self.conc, conc_depend=conc_depend,
                                paired_ranges=paired_ranges_, mc_vector=mc_vector,
                                conc_factor=self.conc_factor,
                                ecurve=self.ecurve, j0=self.j0,
                                )

        self.logger.info('pno=%d, nth=%d, cdl_list=%s, use_elution_models=%s', pno, nth, str(cmatrix.cdl_list), str(self.use_elution_models))
        if not self.use_elution_models:
            if nth > 0:
                nth = 0
                self.logger.info("nth has been changed to %d as a temporary fix in non-elution-model mode.", nth)

        C_ = cmatrix.data
        if save_cmatrix:
            np.savetxt("C-%d-%d.dat" % (start, stop), C_)
        self.ret_C = C_

        D = self.data[:,start:stop]
        E = self.error[:,start:stop]

        rank = C_.shape[0] + surplus_rank
        if not rank_control:
            if get_setting('allow_rank_variation'):
                rank_increment = get_setting('rank_increment')
                rank += rank_increment
                self.logger.info('rank incremented by %d to %d.' % (rank_increment, rank))

        if rank_control:
            svd_reconstruct = 1
        else:
            svd_reconstruct = get_setting('svd_reconstruct')

        j0 = self.j0
        synthesized_solution = False
        boundary = None
        if svd_reconstruct:
            if self.synthesized_lrf and rank > 1 and conc_depend > 1:
                if rank == 2 and conc_depend == 2:
                    from Theory.Rg import compute_corrected_Rg
                    from Theory.SynthesizedLRF import synthesized_lrf, get_reduced_conc
                    Rg, gf, gt, _ = compute_corrected_Rg(self.sd, self.ecurve, pno, self.qv, D, E, range_=(start, stop))
                    Cred = get_reduced_conc(C_, cmatrix.cdl_list)
                    red_rank = len(Cred)
                    P, opt_info, D_, boundary = synthesized_lrf(self.qv, D, C_, rank, Cred, Rg)
                    synthesized_solution = True
                else:
                    from Rank.Synthesized import get_reduced_rank, synthesized_data
                    D_, boundary  = synthesized_data(self.qv, D, self.Rg, rank=rank, cd=conc_depend)
                    red_rank = get_reduced_rank(rank, conc_depend)
                rank_phrase = 'with ranks=(%d, %d)' % (rank, red_rank)
                by_phrase = ' by synthesized LRF'
            else:
                D_ = get_denoised_data(D, rank=rank)
                rank_phrase = 'with rank=%d' % rank
                by_phrase = ''
            self.logger.info('SVD-reconstructed %s in range [%d, %d)%s.', rank_phrase, j0+start, j0+stop, by_phrase)

        else:
            D_ = D
            self.logger.info('skipping SVD-reconstruction in range [%d, %d).', j0+start, j0+stop)

        weight_matrix_type = get_setting('weight_matrix_type')
        if weight_matrix_type == 1:
            from .WeightedLRF import compute_reciprocally_weighted_matrices
            Cr, Dr = compute_reciprocally_weighted_matrices(self.conc_dependence, C_, D_)
            self.logger.info("using concentration-reciprocal weights.")
            if DEBUG_WEIGHT_MATRIX:
                from molass_legacy.DataStructure.MatrixData import simple_plot_3d
                fig = plt.figure(figsize=(21,7))
                ax1 = fig.add_subplot(131, projection='3d')
                ax2 = fig.add_subplot(132, projection='3d')
                ax3 = fig.add_subplot(133)
                simple_plot_3d(ax1, D_)
                simple_plot_3d(ax2, Dr)
                ax3.plot(Cr[0,:])
                ax3.plot(Cr[1,:])
                fig.tight_layout()
                plt.show()
            C_, D_ = Cr, Dr

        self.C_ = C_

        if synthesized_solution:
            pass
        else:
            try:
                Cpinv = np.linalg.pinv( C_ )
            except Exception as exc:
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(self.logger, "pinv failed: ")
                # np.savetxt("C-%d-%d.dat" % (start, stop), C_)
                raise exc

            if ERROR_PROPAGATION_DENOISE:
                E_  = get_denoised_error( D_, D, E )

            self.D  = D_
            self.P_init = np.dot( D_, Cpinv )

            if demo_info is not None:
                self.demo_data = []
                # self.P_init[:,0] += np.random.uniform(size=D.shape[0]) * 0.01
                self.P_init[demo_info[0],0] += 0.002
                self.P_init[demo_info[1],0] -= 0.002
                # self.demo_data.append(self.P_init[demo_info,0])

            known_log_str = '' if known_peak_info is None else ' and known ' + str(known_peak_info)
            self.logger.info( 'solving rank %d factorization of M into P, C with shapes %s = %s @ %s%s'
                                 % ( rank, str(self.D.shape), str(self.P_init.shape), str(C_.shape), known_log_str ) )

            self.set_known_peak_info(known_peak_info, conc_depend)

            P, opt_info = self.fit()

        opt_info += [(nth, cmatrix.cdl_list), self.qv, boundary, D_]

        if self.ignore_bq:
            A   = P[:,nth]
            B   = np.zeros(P.shape[0])
            Z   = B
        else:
            A   = P[:,nth*conc_depend]
            B   = P[:,nth*conc_depend+1]
            if conc_depend > 2:
                Z   = P[:,nth*conc_depend+2]
            else:
                Z   = np.zeros(P.shape[0])

        if self.concentration_datatype < 2:
            j = paired_ranges_[nth].top_x
            top_c = self.mc_vector[j]
            top_y = self.ecurve.y[j]/top_c
            i = self.sd.xray_index
            ab_scale = top_y/A[i]
            A *= ab_scale
            B *= ab_scale**2    # ok?
            Z *= ab_scale**3    # ok?

        if ERROR_PROPAGATION_DENOISE:
            Ae, Be, Ze = self.compute_propagated_errors_from_denoised_error( P, D_, E_, nth, self.ignore_bq )
        else:
            Ae, Be, Ze = self.compute_propagated_errors( P, D, E, nth, self.ignore_bq )

        if self.concentration_datatype < 2:
            Ae *= ab_scale
            Be *= ab_scale**2   # ok?
            Ze *= ab_scale**3   # ok?

        if DEBUG_LRF_RANK:
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
            C = self.ret_C
            M = P@C
            svd_results = []
            s_list = []
            for matrix in [P, C, M]:
                U, s, VT = np.linalg.svd(matrix)
                svd_results.append(s)
                s_list.append(s[0])

            ymax = np.max(s_list)*1.1
            for s, ax, name in zip(svd_results, axes, ['$P$', '$C$', r'$P \cdot C$']):
                ax.set_title('Singular Values of ' + name, fontsize=30)
                ax.plot(s[0:5], ':', marker='o', markersize=10)
                w = np.where(s>0.2)[0]
                print('s=', s)
                print('w=', w)
                rank = len(w)
                ax.text(2, ymax*0.7, name, fontsize=100, alpha=0.2, ha='center', va='center')
                ax.text(2, ymax*0.4, "Rank %d" % rank, fontsize=100, alpha=0.2, ha='center', va='center')
                ax.set_xlim(-0.5, 4.5)
                ax.set_ylim(-0.1, ymax)
            fig.tight_layout()
            plt.show()

        if self.debugger is not None:
            from collections import OrderedDict
            param_info = OrderedDict()
            param_info['rank'] = rank
            param_info['popts'] = str(self.popts)
            param_info['sd_id_info'] = self.sd_id_info
            self.debugger.save_info(D, C_, A, param_info)

        lrfE = [Ae, Be, Ze]
        return A, B, Z, lrfE, LrfInfo(opt_info, A, lrfE), self.ret_C

    def set_known_peak_info(self, known_peak_info, conc_depend):
        if known_peak_info is None:
            self.solver_known_info = known_peak_info
        else:
            self.solver_known_info = []
            for k, info in enumerate(known_peak_info):
                if info is None:
                    data = None
                else:
                    data = info.data
                    n = k*conc_depend
                    self.P_init[:,n] = data[:,1]
                self.solver_known_info.append(data)

    def get_anim_data( self ):
        return self.anim_data1, self.anim_data2

    def compute_propagated_errors_from_denoised_error( self, P, D_, E_, nth, ignore_bq ):
        D_pinv = np.linalg.pinv(D_)
        W   = np.dot(D_pinv, P)
        Pe  = np.sqrt( np.dot(E_**2, W**2) )
        if ignore_bq:
            Ae  = Pe[:,nth]
            Be  = np.zeros(P.shape[0])
            Ce  = Be
        else:
            Ae  = Pe[:,nth*self.conc_depend]
            Be  = Pe[:,nth*self.conc_depend+1]
            if self.conc_depend < 3:
                Ce  = np.zeros(P.shape[0])
            else:
                Ce  = Pe[:,nth*self.conc_depend+2]
        return Ae, Be, Ce

    def compute_propagated_errors( self, P, M, E, nth, ignore_bq ):
        M_pinv = np.linalg.pinv(M)
        W   = np.dot(M_pinv, P)
        Pe  = np.sqrt( np.dot(E**2, W**2) )
        if ignore_bq:
            Ae  = Pe[:,nth]
            Be  = np.zeros(P.shape[0])
            Ce  = Be
        else:
            Ae  = Pe[:,nth*self.conc_depend]
            Be  = Pe[:,nth*self.conc_depend+1]
            if self.conc_depend < 3:
                Ce  = np.zeros(P.shape[0])
            else:
                Ce  = Pe[:,nth*self.conc_depend+2]
        return Ae, Be, Ce

    def compute_error( self, P_flat ):
        if self.stop_event is not None and self.stop_event.is_set():
            raise Exception( 'Stop Event is set' )

        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C_ )

        # compute a squared Frobenius norms
        rec_error = np.linalg.norm( reconstructed - self.D )**2

        if self.aq_positivity:
            A   = np.dot( P, self.P2A )
            negA = np.min( [ self.Zero, A ], axis=0 )
            neg_error = np.linalg.norm( negA )**2
        else:
            neg_error = 0

        if self.add_conc_const:
            base_error = np.linalg.norm( np.dot( P, self.P2E ) )**BASE_ERROR_NORM_DEGREE
        else:
            base_error = 0

        if self.solver_known_info is None:
            known_error = 0
        else:
            known_error = self.compute_known_error(P)

        if self.animation:
            self.anim_counter += 1
            anim_rec1 = [ self.anim_counter, P ]
            anim_rec2 = [ rec_error,
                            neg_error*self.penalty_weights[0],
                            base_error*self.penalty_weights[1] ]

        if self.demo_info is not None:
            self.demo_data.append(P[self.demo_info,0])

        # print( 'ratio=', base_error/rec_error )
        error   = rec_error + neg_error*self.penalty_weights[0] + base_error*self.penalty_weights[1] + known_error

        if self.aq_smoothness:
            if self.ignore_bq:
                penalties = self.sp.get_penalties_bq_ignore(P_flat, self.penalty_weights[2])
                error += np.sum(penalties)
                if self.animation:
                    aq_penalty_sum = np.sum(penalties)
                    anim_rec2 += [ aq_penalty_sum, 0 ]
            else:
                penalties = self.sp.get_penalties(P_flat, self.penalty_weights[2:])
                error += np.sum(penalties)
                if self.animation:
                    aq_penalty_sum = np.sum( penalties[0::2] )
                    bq_panalty_sum = np.sum( penalties[1::2] )
                    anim_rec2 += [ aq_penalty_sum, bq_panalty_sum ]

        if self.animation:
            self.last_anim_rec1 = anim_rec1
            self.last_anim_rec2 = anim_rec2
            nth = max(0, int(np.log10(self.anim_counter+0.5))-1)
            self.anim_pick_data = self.anim_counter % (10 ** nth) == 0
            if self.anim_pick_data:
                self.anim_data1.append( anim_rec1 )
                self.anim_data2.append( anim_rec2 )

        return error

    def compute_known_error(self, P):
        return compute_known_error_impl(self.conc_depend, self.solver_known_info, P)

    def compute_error_grad( self, P_flat ):
        P = P_flat.reshape( self.P_init.shape )
        reconstructed = np.dot( P, self.C_ )

        # compute the derivative of the squared Frobenius norm
        # || P@C - D ||**2 ==> 2*( P@C - D )@C.T
        rec_grad = 2 * np.dot( reconstructed - self.D, self.C_.T ).flatten()

        if self.aq_positivity:
            A   = np.dot( P, self.P2A )
            negA = np.min( [ self.Zero, A ], axis=0 )
            neg_grad = 2 * negA.flatten()
        else:
            neg_grad = 0

        if self.add_conc_const:
            base_grad = 2 * np.dot( np.dot( P, self.P2E ), self.P2E.T ).flatten()
        else:
            base_grad = 0

        if self.solver_known_info is None:
            known_error_grad = 0
        else:
            known_error_grad = self.compute_known_error_grad(P)

        error_grad  = rec_grad + neg_grad * self.penalty_weights[0] + base_grad * self.penalty_weights[1] + known_error_grad

        if self.aq_smoothness:
            if self.ignore_bq:
                error_grad += self.sp.get_penalty_diff_bq_ignore(P_flat, self.penalty_weights[2])
            else:
                error_grad += self.sp.get_penalty_diff(P_flat, self.penalty_weights[2:])

        return error_grad

    def compute_known_error_grad(self, P):
        return compute_known_error_grad_impl(self.conc_depend, self.solver_known_info, P)

    def fit( self ):
        t0 = time.time()
        # return self.P_init

        if self.aq_smoothness:
            from SmoothnessPenalty import SmoothnessPenalty
            self.sp = SmoothnessPenalty(self.P_init.shape[0], self.conc_depend, add_const=self.add_conc_const)

        self.Zero = np.zeros( self.P_init.shape )
        n = self.P_init.shape[1]
        self.P2A = np.zeros( (n, n) )
        for i in range(0, n, 2):
            self.P2A[i,i] = 1

        self.P2E = np.zeros( (n, n) )
        self.P2E[n-1,n-1] = 1

        self.num_iterations = 0
        def num_iterations_count(x):
            self.num_iterations += 1

        ret = optimize.fmin_cg( self.compute_error, self.P_init.flatten(), fprime=self.compute_error_grad,
                maxiter=MAX_ITERATION_FMIN_CG,
                full_output=True,
                callback=num_iterations_count )
        P_flat = ret[0]
        func_calls = ret[2]
        self.P_fit = P = P_flat.reshape( self.P_init.shape )

        if self.animation:
            self.anim_data1.append( self.last_anim_rec1 )
            self.anim_data2.append( self.last_anim_rec2 )

        t1 = time.time()
        time_elapsed = t1 - t0
        print( 'It took ', time_elapsed, 'num_iterations=', self.num_iterations )

        self.logger.info('solved with aq_smoothness=%s, aq_positivity=%s resulting into iterations=%d',
                            str(self.aq_smoothness), str(self.aq_positivity), self.num_iterations)

        return P, [self.num_iterations, func_calls, time_elapsed,]

def run_extrapolation_solver_impl(
        solver, selector, conc_depend=None,
        exe_queue=None, stop_event=None ):

    ignore_all_bqs = get_setting('ignore_all_bqs')
    if DISABLE_IGNORE_BQ_LIST:
        ignore_bq_list = None
    else:
        ignore_bq_list = get_setting('ignore_bq_list')

    results = []
    peak_range_infos = []
    row = 0
    if exe_queue is not None:
        exe_queue.put( [0, row ] )
        if stop_event.is_set():
            return

    # previous_penalty_matrix = get_setting('zx_penalty_matrix')
    # TODO: re-consider on previous_penalty_matrix
    previous_penalty_matrix = None

    for pno, paired_range in enumerate(solver.cnv_ranges):
        fromto_list = paired_range.get_fromto_list()
        for rno, range_ in enumerate(fromto_list):
            start   = range_[0]
            stop    = range_[1] + 1
            peakset_info = selector.select_peakset(row)
            penalty_weights = None if previous_penalty_matrix is None else previous_penalty_matrix[row]
            if ignore_bq_list is None:
                ignore_bq = ignore_all_bqs
            else:
                ignore_bq = ignore_bq_list[row]
            try:
                A, B, Z, E, lrf_info, C = solver.extrapolate_wiser( start, stop, peakset_info,
                                        stop_event=stop_event,
                                        penalty_weights=penalty_weights,
                                        ignore_bq=ignore_bq,
                                        conc_depend=conc_depend,
                                        )
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb)
                if exe_queue is not None:
                    exe_queue.put( [-1, str(etb) ] )
                raise RuntimeError('in the solver')

            if stop_event is not None and stop_event.is_set():
                print( "solver stopped by stop_event." )
                return None

            results.append( [A, B, Z, E, lrf_info, C] )
            peak_range_infos.append( [pno, rno, paired_range] )
            row += 1
            if exe_queue is not None:
                exe_queue.put( [0, row ] )

    print( "solver finished." )
    return results, peak_range_infos

def run_extrapolation_solver_with_progress(
        solver, selector, parent, conc_depend=None):

    import queue
    import threading
    from ProgressMinDialog import ProgressMinDialog
    from molass_legacy.KekLib.TkUtils import split_geometry

    w, h, x, y = split_geometry( parent.winfo_geometry() )
    geometry_info = "+%d+%d" % (parent.winfo_rootx() + int(w*0.7), parent.winfo_rooty() + int(h*0.8))

    exe_queue = queue.Queue()
    max_iter = len( selector.required_peakset_infos ) + 1

    stop_event = threading.Event()
    def proc():
        ret_tuple = run_extrapolation_solver_impl(
            solver, selector,conc_depend=conc_depend,
            exe_queue=exe_queue, stop_event=stop_event )
        exe_queue.put( [1, ret_tuple] )

    thread = threading.Thread( target=proc, name='PreviewExtrapolationgThread', args=[] )
    thread.start()

    progress_dialog = [None]

    counter = [0]
    ret_info = [None]
    def progress_cb():
        error_info = None
        try:
            info = exe_queue.get(block=False)
            print('progress_cb:', info[0])
            if info[0] == 1:
                ret_info[0] = info[1]
                counter[0] = max_iter + 1
            elif info[0] == 0:
                counter[0] = info[1] + 1
            else:
                # get exception info
                error_info = info[1]
        except:
            pass

        if error_info is not None:
            pdialog = progress_dialog[0]
            pdialog.cancel()
            thread.join()
            # raise the exception here in the parent thread
            # which occrured in the child thread
            raise RuntimeError(error_info)
            """
                this will be caught in Tk with the overriden repoter
                GuiMain.report_callback_exception if this code is running
                under GuiMain
            """

        return counter[0]

    progress = ProgressMinDialog(parent,
                    title="Preview Extrapolation with Smoothness Constraints",
                    message="It may take several minutes. Please be patient.",
                    num_steps=max_iter + 1, length=360,
                    progress_cb=progress_cb, geometry_info=geometry_info,
                    cancelable=True, visible=False )

    progress_dialog[0] = progress
    progress.show()

    if progress.canceled:
        stop_event.set()

    thread.join()

    ret_tuple = ret_info[0]
    return ret_tuple

def run_extrapolation_solver(
        solver, selector, conc_depend=None, parent=None, use_gui=True ):

    if parent is None or not use_gui:
        ret_tuple = run_extrapolation_solver_impl(
            solver, selector, conc_depend=conc_depend)
    else:
        ret_tuple = run_extrapolation_solver_with_progress(
            solver, selector, parent, conc_depend=conc_depend)

    return ret_tuple



"""

    DevelUtils.LegacyImporters.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""

def import_legacy_modules():
    import molass_legacy.KekLib, SerialAnalyzer, DataStructure, GuinierAnalyzer, Decomposer, Extrapolation



# coding: utf-8
"""
    ErHypothesis.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
from bisect import bisect_right
import numpy as np
from scipy.optimize import minimize
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Rectangle
from matplotlib.widgets import Button
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from molass_legacy.Models.ElutionCurveModels import emg_x_from_height_ratio
from SvdDenoise import get_denoised_data

class ErHypothesisDialog(Dialog):
    def __init__( self, parent, seed=None, noise=0.01):
        self.init_seed = seed
        self.noise = noise
        Dialog.__init__( self, parent, "ErHypothesis", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(21, 10))
        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        gs = GridSpec(2,3)
        ax00 = fig.add_subplot(gs[0,0], projection='3d')
        ax01 = fig.add_subplot(gs[0,1])
        ax02 = fig.add_subplot(gs[0,2])
        ax10 = fig.add_subplot(gs[1,0])
        ax11 = fig.add_subplot(gs[1,1])
        ax12 = fig.add_subplot(gs[1,2])
        self.axes = np.array([[ax00, ax01, ax02], [ax10, ax11, ax12]])

        self.generate_data(modeled=False, seed=self.init_seed)

        fig.suptitle("Verification Simulation for Exact Rank Hypothesis; random seed=%d" % self.seed, fontsize=30)

        self.draw_3d(ax00)
        self.draw_sigmas(ax10)
        self.draw_input_curves(ax01, ax11)
        self.draw_p_curves(ax02, ax12)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9, bottom=0.1)
        self.mpl_canvas.draw()

    def buttonbox( self ):
        bottom_frame = Tk.Frame(self)
        bottom_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.05

        tframe = Tk.Frame(bottom_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(bottom_frame, width=width*0.25)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(bottom_frame)
        box.pack(side=Tk.RIGHT)

        self.almerge = Tk.IntVar()
        self.almerge.set(0)
        w = Tk.Checkbutton(box, text="ALMERGE", variable=self.almerge)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        w = Tk.Button(box, text="Solve", width=10, command=self.solve_several_ways, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=10, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def generate_data(self, modeled=True, seed=None, noise=None):
        if seed is None:
            seed = np.random.randint(1000, 9999)

        if noise is None:
            noise = self.noise

        if modeled:
            num_elutions = 300
            self.qvector = qvector = np.linspace(0.007, 0.5, 600)
            self.md = md = ModeledData(qvector, num_elutions)
            self.M = M = md.get_data(noise=noise)
            h, mu, sigma, tau = self.md.e_params[0]
            self.conc = 0
        else:
            from .SimData import SimData
            in_folder = 'D:/PyTools/reports/analysis-erh/preview_results'
            a_file = in_folder + '/A1.dat'
            b_file = in_folder + '/B1.dat'
            simd = SimData(a_file, b_file)
            self.qvector = qvector = simd.q
            conc = 1
            M, E = simd.get_data(noise=noise, conc=conc, seed=seed)
            self.M = M
            self.E = E
            h, mu, sigma, tau = simd.e_params[0]
            self.conc = conc
            self.truth = simd.a

        self.seed = seed
        self.i = i = bisect_right(qvector, 0.02)
        self.ez = z =  M[i,:]
        c = z/np.max(z)
        self.C = np.array([c, c**2])

        f, t = [int(p+0.5) for p in emg_x_from_height_ratio(0.5, mu, sigma, tau)]
        print("(f, t)=", (f, t))
        self.eslice = eslice = slice(f,t+1)

        U, s, VT = np.linalg.svd(self.M[:,eslice])
        self.sigmas = s
        rank = 2
        M_ = U[:,0:rank] @ np.diag(s[0:rank]) @ VT[0:rank,:]
        C_ = self.C[:,eslice]
        self.P = M_ @ np.linalg.pinv(C_)

    def draw_3d(self, ax):
        qvector = self.qvector
        M = self.M
        num_elutions = M.shape[1]

        simple_plot_3d(ax, M, x=qvector)

        x = np.ones(num_elutions)*qvector[self.i]
        y = np.arange(num_elutions)
        z = self.ez
        ax.plot(x, y, z, color='orange')
        self.elution = z

        self.j = j = np.argmax(z)
        x = qvector
        y = np.ones(len(x))*j
        z = M[:,j]
        ax.plot(x, y, z, color='green')
        self.ridge = z

    def draw_sigmas(self, ax):
        s = self.sigmas
        n = min(5, len(s))
        ax.plot(np.arange(n), s[0:n], ':', marker='o')
        tx = np.average(ax.get_xlim())
        ty = np.average(ax.get_ylim())
        ax.text(tx, ty, r'$\frac{\sigma_1}{\sigma_0}=%.3g$' % (s[1]/s[0]), alpha=0.5, fontsize=40, ha='center', va='center')

    def draw_input_curves(self, axe, axs):
        axe.plot(self.elution, color='orange')
        f = self.eslice.start
        t = self.eslice.stop-1
        ymin, ymax = axe.get_ylim()
        p = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        axe.add_patch(p)

        axs.set_yscale('log')
        axs.plot(self.qvector, self.ridge)

    def draw_p_curves(self, ax, axt):
        P = self.P
        a = P[:,0]
        b = P[:,1]

        ax.plot(self.qvector, a, color='C1', label='$A_2$')
        for ax_ in [ax, axt]:
            ax_.plot(self.qvector, b, color='pink', label='$B_2$')

    def solve_several_ways(self):
        import molass_legacy.KekLib.DebugPlot as dplt

        if self.conc < 0.5:
            rank_pairs = [(1, 1), (2, 1), (2, 2), (3, 2), (0, 1), (0, 2)]
        else:
            rank_pairs = [(2, 2), (2, 1), (3, 2), (4, 2), (0, 1), (0, 2)]

        scores = []
        for m, n in rank_pairs:
            score = self.evaluate_rank_mn_solution(m, n)
            scores.append(score)

        if self.almerge.get():
            score = self.evaluate_almerge_solution()
            scores.append(score)

        dplt.set_global_opts(kill_button=False)     # reset later if necessary
        dplt.push()
        fig, ax = dplt.subplots()

        ax.set_title("Evaluation of Various Rank (m,n) Solutions", fontsize=20)
        ax.set_ylabel("Norm of Deviation from true Curve", fontsize=16)
        ax.set_yscale('log')

        for k, score in enumerate(scores):
            label = 'Rank %s' % str(rank_pairs[k]) if k < len(rank_pairs) else 'ALMERGE'
            ax.plot(k, score, 'o', label=label)
            if k == 0:
                dx = 0.1
                dy = 2
                ax.annotate("Exact Rank", xy=(k, score), xytext=(k+dx, score*dy),
                                arrowprops=dict(arrowstyle='->', color='k'))

        xmin, xmax = ax.get_xlim()
        ax.set_xlim(xmin, xmax)
        score = scores[0]
        ax.plot([xmin, xmax], [score, score], ':', color='red', alpha=0.5)

        ax.legend()
        fig.tight_layout()
        dplt.show()
        dplt.pop()

    def evaluate_rank_mn_solution(self, m, n):
        assert n in [1, 2]
        eslice = self.eslice
        M_ = get_denoised_data(self.M[:,eslice], rank=m)
        if n == 1:
            C_ = self.C[0,eslice][np.newaxis,:]
        else:
            C_ = self.C[:,eslice]
        P = M_ @ np.linalg.pinv(C_)
        solution = P[:,0]

        def obj_func(pv):
            return np.sum((solution*pv[0] - self.truth)**2)

        res = minimize(obj_func, (1,))
        return res.fun

    def evaluate_almerge_solution(self):
        from ReAtsas.Almerge import Almerge
        from molass_legacy._MOLASS.SerialSettings import get_setting

        eslice = self.eslice
        c = self.C[0,:]
        datafiles = self.save_data()
        almerge = Almerge()
        indeces = np.arange(eslice.start, eslice.stop)
        out_file = get_setting('temp_folder') + '/extrapolated.dat'
        ret = almerge.execute(c, datafiles, indeces, out_file)
        solution = ret.exz_array[:,1]

        def obj_func(pv):
            return np.sum((solution*pv[0] - self.truth)**2)

        res = minimize(obj_func, (1,))
        return res.fun

    def save_data(self, out_folder=None):
        from DataUtils import get_pytools_folder
        from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
        from molass_legacy.KekLib.NumpyUtils import np_savetxt

        if out_folder is None:
            out_folder = get_pytools_folder() + '/temp'

        clear_dirs_with_retry([out_folder])
        datafiles = []
        qv = self.qvector
        for j in range(self.M.shape[1]):
            file = 'simdata_%05d.dat' % j
            path = '/'.join([out_folder, file])
            np_savetxt(path, np.array([qv, self.M[:,j], self.E[:,j]]).T)
            datafiles.append(path)
        return datafiles



"""
    ATSAS.DatCmp.py

    adapted from
        ATSAS.CorMapAnalysis.plot_heatmap(...)

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import re
import numpy as np
import glob
import subprocess as sp
from .AtsasUtils import get_atsas_bin_path

datcmp_exe = os.path.join(get_atsas_bin_path(), "datcmp.exe")

def run_datcmp(folder=None, files=None, return_dict=False, P_threshold=0.01):
    if files is None:
        assert folder is not None
        files = glob.glob(folder + r"\*")

    ret = sp.run([datcmp_exe] + files, capture_output=True)

    datcmp_data = get_dict_from_log(ret.stdout.decode())

    num_frames = len(files)
    full_data = get_full_data_from_dict(num_frames, datcmp_data, P_threshold=P_threshold)

    if return_dict:
        return full_data, datcmp_data
    else:
        return full_data

def get_dict_from_log(log):
    # borrowed from CorMapAnalysis.py
    data_dict = {"1,2": 0}
    for line in iter(log.splitlines()):
        match_obj = re.match(r'\s* \d{1,} vs', line)
        if match_obj:
            data = line.split()
            if "*" in data[5]:
                data[5] = data[5][:-1]
            data_dict["{},{}".format(data[0], data[2])] = [int(float(data[3])),
                                                           float(data[4]),
                                                           float(data[5])]
    return data_dict

def get_full_data_from_dict(num_frames, datcmp_data, P_threshold=0.01, use_adjP=True):

    if use_adjP:
        P_col = 3
    else:
        P_col = 2

    x_axis = np.linspace(1, num_frames, num_frames)

    full_data = []
    for frame in range(1, num_frames+1):
        pw_data = np.zeros([num_frames, 3])
        for i in range(num_frames):
            if i+1 < frame:
                key = "{},{}".format(i+1, frame)
            elif i+1 > frame:
                key = "{},{}".format(frame, i+1)
            else:
                continue
            pw_data[i, :] = np.asarray(datcmp_data[key])

        pw_data = np.column_stack([x_axis, pw_data])
        good_points = pw_data[pw_data[:, P_col] == 1]
        ok_points = pw_data[np.logical_and(1 > pw_data[:, P_col], pw_data[:, P_col] >= P_threshold)]
        bad_points = pw_data[pw_data[:, P_col] < P_threshold]

        xOrder = list(good_points[:, 0]) + list(ok_points[:, 0]) + list(bad_points[:, 0])
        C_values = list(good_points[:, 1]) + list(ok_points[:, 1]) + list(bad_points[:, 1])
        xData = [-1]*len(good_points[:, 0]) + [0]*len(ok_points[:, 0]) + [1]*len(bad_points[:, 0])

        xOrder_sorted, xData_sorted = (list(t) for t in zip(*sorted(zip(xOrder, xData))))
        full_data.append(xData_sorted)

    ret_array = np.asarray(full_data)
    di = np.diag_indices(num_frames)
    ret_array[di] = 0   # set ok values to the diagonal
    return ret_array

def run_datcmp_from_array(qv, M, E, temp_dir=None, P_threshold=0.01, keep_temp=False):
    import shutil
    if temp_dir is None:
        temp_dir = "datcmp_temp"
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)
    os.makedirs(temp_dir)
    for j in range(M.shape[1]):
        file = os.path.join(temp_dir, "data-%04d.dat" % j)
        np.savetxt(file, np.vstack([qv, M[:,j], E[:,j]]).T)
    ret = run_datcmp(temp_dir, P_threshold=P_threshold)
    if not keep_temp:
        shutil.rmtree(temp_dir)
    return ret



"""
    BaselineUtils.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Mapping.MapperConstructor import create_mapper

def get_default_baseline_type():
    xray_baseline_type = get_setting("xray_baseline_type")
    ret_type = 1
    if xray_baseline_type == 5:
        ret_type = 2
    return ret_type

def get_corrected_sd_impl(sd, sd_orig, pre_recog, mapped_info=None):
    # xrav correction
    if mapped_info is None:
        mapper = create_mapper(None, sd, sd_orig, pre_recog)
        mapped_info = mapper.get_mapped_info()
    sd.apply_baseline_correction(mapped_info, basic_lpm=True)   # see Mapping.ElutionMapper.compute_xray_baseline()

    # uv correction
    # not required for v2 optimization

    return sd

def create_xr_baseline_object(unified_baseline_type=None):
    if unified_baseline_type is None:
        unified_baseline_type = get_setting("unified_baseline_type")

    if unified_baseline_type == 1:
        from .LinearBaseline import LinearBaseline
        return LinearBaseline()
    elif unified_baseline_type == 2:
        from .IntegralBaseline import IntegralBaseline
        return IntegralBaseline()
    elif unified_baseline_type == 3:
        from .FoulingBaseline import FoulingBaseline
        return FoulingBaseline()
    else:
        # currently not supported
        assert False

def demo(in_folder, logger):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at
    from MatrixData import simple_plot_3d
    from molass_legacy.Elution.CurveUtils import simple_plot
    from DataUtils import get_in_folder

    set_setting("in_folder", in_folder)
    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd)
    sd_copy = sd.get_copy()
    corrected = get_corrected_sd_impl(sd_copy, sd, pre_recog)

    D0, E0, qv0, xr_curve0 = sd.get_xr_data_separate_ly()
    U0,  _, wv0, uv_curve0 = sd.get_uv_data_separate_ly()
    uv_curve0_2 = pre_recog.flowchange.a_curve2

    D1, E1, qv1, xr_curve1 = corrected.get_xr_data_separate_ly()
    U1,  _, wv1, uv_curve1 = corrected.get_uv_data_separate_ly()
    uv_curve1_2 = make_secondary_e_curve_at(corrected.absorbance.data, wv1, uv_curve1, logger)

    with plt.Dp():
        in_folder_ = get_in_folder()
        fig = plt.figure(figsize=(20,10))
        fig.suptitle("Correction of %s" % in_folder_, fontsize=20)
        ax1 = fig.add_subplot(241, projection="3d")
        ax2 = fig.add_subplot(242, projection="3d")
        ax3 = fig.add_subplot(243)
        ax4 = fig.add_subplot(244)
        ax5 = fig.add_subplot(245, projection="3d")
        ax6 = fig.add_subplot(246, projection="3d")
        ax7 = fig.add_subplot(247)
        ax8 = fig.add_subplot(248)
        ax1.set_title("Input", fontsize=16)
        ax2.set_title("Corrected", fontsize=16)
        ax3.set_title("Input", fontsize=16)
        ax4.set_title("Corrected", fontsize=16)

        simple_plot_3d(ax1, D0, x=qv0)
        simple_plot_3d(ax2, D1, x=qv1)
        simple_plot(ax3, xr_curve0, color="orange")
        simple_plot(ax4, xr_curve1, color="orange")
        simple_plot_3d(ax5, U0, x=wv0)
        simple_plot_3d(ax6, U1, x=wv1)
        simple_plot(ax7, uv_curve0, color="blue")
        simple_plot(ax8, uv_curve1, color="blue")
        ax7t = ax7.twinx()
        ax8t = ax8.twinx()
        ax7t.grid(False)
        ax8t.grid(False)
        def plot_curve_sub(ax, curve):
            x = curve.x
            y = curve.y
            ax.plot(x, y, alpha=0.5)
        plot_curve_sub(ax7t, uv_curve0_2)
        plot_curve_sub(ax8t, uv_curve1_2)

        fig.tight_layout()
        plt.show()



"""
    ExcelCOM.py - a minimal Python wrapper for Excel ole object

    Copyright (c) 2016-2025, Masatsuyo Takahashi, KEK-PF
"""
import os
import re
import time
import numpy as np
from win32com.client import Dispatch, DispatchEx, gencache
from win32process import GetWindowThreadProcessId
import pythoncom
from shutil import copyfile
import logging
import psutil

DEBUG = False

WAIT_TIME_BEFORE_KILL           = 10

msoTrue                         = -1
msoFalse                        =  0
msoConnectorStraight            =  1
msoShapeOval                    =  9
msoShapeRightBrace              = 32
msoTextOrientationHorizontal    =  1
msoThemeColorAccent1            = 5
msoThemeColorText1              = 13
xlCategory                      = 1
xlValue                         = 2
xlLine                          = 4
xlUpward                        = -4171
xlLow                           = -4134
xlNone                          = -4142
xlXYScatter                     = -4169
xlMaximum                       = 2

worksheet_path_re = re.compile('^(.+\.xls[xm])\((.+)\)$')

def CoInitialize():
    pythoncom.CoInitialize()

def CoUninitialize():
    pythoncom.CoUninitialize()

def ColAddr( col ):
    col_ = col - 1
    addr = ''
    while True:
        r = col_ % 26
        col_ //= 26
        addr = chr(65+r) + addr
        if col_ == 0:
            break
        else:
            col_ -= 1
    return addr

def CellAddr( cell, absref='$$' ):
    row_abs = absref[0].replace( ' ', '' )
    col_cabs = absref[1].replace( ' ', '' )
    return row_abs + ColAddr( cell[1] ) + col_cabs + str(cell[0])

def RGB( hex_color_str ):
    return int( hex_color_str[4:6] + hex_color_str[2:4] + hex_color_str[0:2], 16 )

CLSID = 'Excel.Application'
INVOKE_EXCEL_SIMPLE_DISPATCH = True

pid_list = []

def cleanup_created_excels():
    logger = logging.getLogger( __name__  )
    while len(pid_list) > 0:
        pid = pid_list.pop(0)
        try:
            proc    = psutil.Process( pid )
            if proc.is_running():
                proc.kill()
                logger.warning( 'Killed the remaining process pid=' + str(pid) + ' in cleanup.' )
        except Exception as exc:
            # psutil.NoSuchProcess no process found with pid ...
            logger.debug( str(exc) + '; ok' )

class ExcelComClient:
    def __init__( self, new_instance=True, excel_id=None, keep_remaining=False, visible=False, display_alerts=False ):
        self.clsid = None
        self.excel = None
        self.new_instance = None
        self.keep_remaining = keep_remaining
        self.visible = visible
        self.logger = logging.getLogger( __name__  )
        if excel_id is None:
            if new_instance:
                self.excel  = DispatchEx(CLSID)
                self.new_instance = True
            else:
                self.excel  = Dispatch(CLSID)
                self.new_instance = False
        else:
            self.new_instance = False
            self.excel  = Dispatch(
                            pythoncom.CoGetInterfaceAndReleaseStream( excel_id, pythoncom.IID_IDispatch )
                            )

        if self.visible:
            self.excel.Visible = True

        self.excel.DisplayAlerts = display_alerts

        self.pid = GetWindowThreadProcessId( self.excel.Hwnd )[1]
        if self.new_instance:
            pid_list.append( self.pid )
        self.logger.debug( 'Excel started with pid %d' % self.pid )

    def __del__( self ):
        if self.excel is not None:
            # self.logger.info( 'Excel Quit in ExcelComClient.__del__' )
            self.quit()

        if self.new_instance:
            time.sleep(1)
            try:
                proc    = psutil.Process( self.pid )
                n = 0
                while n < WAIT_TIME_BEFORE_KILL and proc.is_running():
                    time.sleep(1)
                    n += 1
                if proc.is_running():
                    if self.keep_remaining:
                        self.logger.warning( 'Kept the remaining process pid=' + str(self.pid) )
                    else:
                        proc.kill()
                        self.logger.warning( 'Killed the remaining process pid=' + str(self.pid) )
            except Exception as exc:
                # psutil.NoSuchProcess no process found with pid ...
                self.logger.debug( str(exc) + '; ok' )

            try:
                pid_list.remove( self.pid )
            except ValueError:
                # this happens in cases after cleanup_created_excels
                pass

    def get_pid( self ):
        self.pid = GetWindowThreadProcessId(self.excel.Hwnd)[1]
        return self.pid

    def get_comid( self ):
        excel_id    = pythoncom.CoMarshalInterThreadInterfaceInStream( pythoncom.IID_IDispatch, self.excel )
        return excel_id

    def get_version( self ):
        return self.excel.Version

    def get_thread_mode(self):
        return self.excel.MultiThreadedCalculation.ThreadMode

    def get_thread_count(self):
        return self.excel.MultiThreadedCalculation.ThreadCount

    def create_thread( self, target, args=[], kwargs={}, name='ExcelComClientThread' ):
        import threading
        thread      = threading.Thread(
                                    target=target,
                                    name=name,
                                    args=args,
                                    kwargs=kwargs
                                    )
        return thread

    def selection( self ):
        return self.excel.Selection

    def quit( self ):
        # self.excel.DisplayAlerts = alert
        try:
            self.excel.Quit()
            self.logger.debug( 'Excel Quit ok' )
        except AttributeError:
            # AttributeError: Excel.Application.Quit
            self.logger.warning( 'Excel with pid %d seems to have died' % self.pid )
        self.excel = None

    def openWorkbook( self, path, save=True ):
        path_ = path.replace( '/', '\\' )
        wb = Workbook_( self, path_, save )
        return wb

    def openWorksheet( self, path ):
        path_ = path.replace( '/', '\\' )
        ws = Worksheet( self, path_ )
        return ws

    def merge_books( self, input_books, merged_book, progress_cb=None, merged_cb=None, delete_target=None, default_font=None ):
        if merged_book in input_books:
            raise Exception( "input_books must not include merged_book" )

        if os.path.exists( merged_book ):
            os.remove( merged_book )

        # Opening a new book causes a loss of book parameters
        # which are not copied by ws.Copy method.
        # So, it is better to filecopy to preserve ones in the
        # first book.
        copyfile( input_books[-1], merged_book )

        if progress_cb is not None:
            progress_cb(0)

        if len( input_books ) == 1:
            return

        out_wb = self.openWorkbook( merged_book ).workbook
        if default_font is not None:
            out_wb.Styles("Normal").Font.Name = default_font

        # num_sheets = out_wb.Sheets.Count
        target_ws = out_wb.Sheets(1)
        # self.logger.info( 'debug: target_ws.Name=' + target_ws.Name )

        for i, book in enumerate(input_books[:-1]):
            try:
                in_wb = self.openWorkbook( book ).workbook
            except:
                self.logger.error( 'Skipped because of an open error: ' + book )
                continue

            for ws in in_wb.Sheets:
                # print( "ws.Name", ws.Name )
                # self.logger.info( 'debug: copying ' + ws.Name )
                # ? After results in a new book
                ws.Copy( Before=target_ws )
            in_wb.Close( SaveChanges=False )
            if progress_cb is not None:
                progress_cb(i)

        if target_ws.Name == delete_target:
            target_ws.Delete()

        if merged_cb is not None:
            merged_cb( self, out_wb )

        out_wb.Sheets(1).Activate()
        out_wb.Save()
        out_wb.Close()

def merge_into_a_book( input_books, merged_book, excel_client=None, progress_cb=None, delete_target=None, default_font=None ):
    if excel_client is None:
        CoInitialize()  # required in thread
        excel_client_ = ExcelComClient()
    else:
        excel_client_ = excel_client

    """
    TODO: this doen't work!
    def freeze_pane_sheet1( excel_client, wb ):
        # workaourd to fix broken FreezePanes setting
        ws = wb.Worksheets(1)
        ws.Activate()
        excel_client.excel.Range( "D2" ).Select()
        aw = excel_client.excel.ActiveWindow
        aw.FreezePanes  = True
        aw.ScrollColumn = 16

    excel_client_.merge_books( input_books, merged_book, progress_cb=progress_cb, merged_cb=freeze_pane_sheet1 )
    """
    excel_client_.merge_books( input_books, merged_book, progress_cb=progress_cb, delete_target=delete_target, default_font=default_font )

    if excel_client is None:
        excel_client_ = None
        CoUninitialize()

"""
    Naming this class as Workbook_ is to distinguish Workbook  from openpyxl
"""
class Workbook_:
    def __init__( self, excel_client, path, save=True ):
        self.excel_client   = excel_client
        self.excel          = excel_client.excel
        if os.path.exists( path ):
            wb_ = self.excel.Workbooks.Open( path )
        else:
            wb_ = self.excel.Workbooks.Add()
            if save:
                wb_.SaveAs( path )

        self.workbook = wb_

class Worksheet:
    def __init__( self, excel_client, path ):
        self.excel_client   = excel_client
        self.excel          = excel_client.excel
        m = worksheet_path_re.match( path )
        if not m: raise Exception('Invalid path: ' + path )

        wb_path = m.group(1)
        ws_name = m.group(2)
        # print( 'Worksheet:', wb_path, ws_name )

        self.workbook   = wb_ = self.excel_client.openWorkbook( wb_path ).workbook
        try:
            ws_ = wb_.Worksheets(ws_name)
        except:
            ws_ = wb_.Worksheets(1)
            ws_.Name = ws_name

        # print( 'ws_.Name=', ws_.Name )
        self.worksheet  = ws_

    def draw_line( self, p1, p2 ):
        line = self.worksheet.Shapes.AddConnector( msoConnectorStraight, p1[0], p1[1], p2[0], p2[1] )
        line.Select()
        return line

    def draw_oval( self, p, a, b, color=None ):
        oval = self.worksheet.Shapes.AddShape( msoShapeOval, p[0], p[1], a, b )
        oval.Select()
        if color is not None:
            selection = self.excel.Selection
            selection.ShapeRange.Fill.ForeColor.RGB = color
        return oval

    def draw_right_brace( self, p, width, hight, rotation ):
        """
        Function AddShape(Type As MsoAutoShapeType, Left As Single, Top As Single, Width As Single, Height As Single) As Shape
        """
        brace = self.worksheet.Shapes.AddShape( msoShapeRightBrace, p[0], p[1], width, hight )
        brace.Select()
        self.excel.Selection.ShapeRange.Rotation = rotation
        return brace

    def draw_textbox( self, text, p, width, hight, visible=True ):
        textbox = self.worksheet.Shapes.AddTextbox( msoTextOrientationHorizontal, p[0], p[1], width, hight )
        textbox.Select()
        selection = self.excel.Selection
        selection.ShapeRange.Line.Visible = visible
        selection.ShapeRange(1).TextFrame2.TextRange.Characters.Text = text
        return textbox

    def freeze_panes( self, range_, scroll_column=None ):
        self.excel.Range( range_ ).Select()
        self.excel.ActiveWindow.FreezePanes = True
        if scroll_column is not None:
            self.excel.ActiveWindow.ScrollColumn = scroll_column

    def zoom( self, percentage ):
        self.excel.ActiveWindow.Zoom = percentage

    def get_num_charts( self ):
        return self.worksheet.ChartObjects().Count

    def get_chart(self, i, debug=False):
        if debug:
            for chart in self.worksheet.ChartObjects():
                print( chart.Name )

        chartobj = self.worksheet.ChartObjects()[i]

        if debug:
            print( 'chartobj.Chart=', chartobj.Chart )
            for ser in chartobj.Chart.SeriesCollection():
                print( ser.Name )

        return chartobj.Chart

    def get_chart_data_point( self, chart, i, j ):
        # <class 'numpy.int64'> can cause 'MemoryError: CreatingSafeArray'
        i_, j_ = int(i), int(j)
        point  = chart.SeriesCollection()[i_].Points()[j_]

        if DEBUG:
            print( 'point.Left=', point.Left )
            print( 'point.Top=', point.Top )

        return point

def compute_axis_max_value( max_value, num_labels=5 ):
    log10_  = int( np.floor( np.log10( max_value / num_labels ) ) )
    scale   = np.power( 10.0, log10_ )
    scaled_value = max_value/scale
    # print( 'scale=', scale, 'scaled_value=', scaled_value )
    return np.ceil( scaled_value ) * scale

def excel_availability(debug=False):
    """
    Note that this test may require rebooting after changing the Excel abailability.
    """
    CoInitialize()
    version = None
    try:
        excel = ExcelComClient()
        version  = excel.get_version()
        excel.quit()
    except:
        if debug:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
    CoUninitialize()
    return version



"""
    DecompEditorDebug.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def debug_plot_impl(frame, dialog):

    x = frame.x
    fx = frame.fx
    y = frame.y
    uv_y = frame.uv_y
    j0 = frame.xr_j0

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5))

        ty = np.zeros(len(x))
        ax1.plot(x, uv_y)
        for k, rec in enumerate(frame.opt_recs_uv):
            # print([k], "str(rec)=", str(rec))
            func = rec[1]
            cy = func(fx)
            ax1.plot(x, cy, ":")
            ty += cy

        ax1.plot(x, ty, ":", color="red", lw=2)

        ax2.plot(x, y)

        ty = np.zeros(len(x))
        for k, rec in enumerate(frame.opt_recs):
            print([k], "str(rec)=", str(rec))
            # func = rec[1]
            func = rec.evaluator
            cy = func(fx, debug=True)
            ax2.plot(x, cy, ":")
            ty += cy

        ax2.plot(x, ty, ":", color="red", lw=2)

        fig.tight_layout()
        plt.show()



"""
    optimizer.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""

def main():
    import os
    import sys
    import getopt
    this_dir = os.path.dirname( os.path.abspath( __file__ ) )
    root_dir = os.path.dirname(os.path.dirname( this_dir ))
    sys.path.insert(0, root_dir)

    optlist, args = getopt.getopt(sys.argv[1:], 'c:w:f:n:i:b:d:m:s:r:t:p:T:M:S:')
    print(optlist, args)
    optdict = dict(optlist)
    if optdict.get('-r'):
        main_impl(optdict, optlist)
    else:
        main_tk(optdict, optlist)

def main_tk(optdict, optlist):
    from molass_legacy.KekLib.TkUtils import get_tk_root
    root = get_tk_root()
    def run_main():
        main_impl(optdict, optlist)
        root.quit()
    root.after(0, run_main)
    root.mainloop()
    root.destroy()

def main_impl(optdict, optlist):
    import os
    import numpy as np
    from molass_legacy.KekLib.ChangeableLogger import Logger
    from molass_legacy._MOLASS.Version import get_version_string
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Optimizer.OptimizerSettings import OptimizerSettings
    from molass_legacy.Optimizer.OptimizerMain import optimizer_main
    from molass_legacy.Optimizer.TheUtils import get_analysis_folder_from_work_folder
    from molass_legacy.Optimizer.SettingsSerializer import unserialize_for_optimizer

    work_folder = optdict['-w']
    os.chdir(work_folder)

    nnn = int(work_folder[-3:])

    log_file = "optimizer.log"
    logger = Logger(log_file)

    logger.info(get_version_string())

    analysis_folder = get_analysis_folder_from_work_folder(work_folder)
    set_setting("analysis_folder", analysis_folder)
    optimizer_folder = os.path.join(analysis_folder, "optimized")
    set_setting("optimizer_folder", optimizer_folder)   # optimizer_folder will be referenced in DataTreatment.load()

    in_folder = optdict['-f']
    set_setting("in_folder", in_folder)     # required in the devel mode of SecTheory.ColumnTypes.py

    settings = OptimizerSettings()
    settings.load(optimizer_folder=optimizer_folder)    # this should restore required temporary settings
    logger.info("optimizer settings restored as %s", str(settings))

    try:
        class_code = optdict['-c']
        n_components = int(optdict['-n'])
        init_params_txt = optdict['-i']
        init_params = np.loadtxt(init_params_txt)
        bounds_txt = optdict['-b']
        if os.path.exists(bounds_txt):
            real_bounds = np.loadtxt(bounds_txt)
        else:
            real_bounds = None
        drift_type = optdict['-d']
        niter = int(optdict['-m'])
        seed = int(optdict['-s'])
        trimming_txt = optdict['-r']
        sleep_seconds = optdict.get('-t')

        unserialize_for_optimizer(optdict.get('-p'))    # "poresize_bounds", "t0_upper_bound"

        test_pattern = optdict.get('-T')
        if test_pattern != "None":
            set_setting("test_pattern", test_pattern)

        with open("in_data_info.txt", "w") as fh:
            fh.write("in_folder=%s\n" % in_folder)

        callback_txt = "callback.txt"
        with open(callback_txt, "w") as fh:
            pass

        with open("pid.txt", "w") as fh:
            fh.write("pid=%d\n" % os.getpid())

        with open("seed.txt", "w") as fh:
            fh.write("seed=%d\n" % seed)

        shm_name = optdict.get('-M')
        if shm_name is None or shm_name  == "None":
            shared_memory = None
        else:
            from molass_legacy.Optimizer.NpSharedMemory import get_shm_proxy
            shared_memory = get_shm_proxy(shm_name)

        solver = optdict.get('-S')

        if sleep_seconds is None:
            logger.info("optimizer started with class_code=%s, optlist=%s, shared_memory=%s", class_code, str(optlist), shm_name)
            optimizer_main(in_folder,
                    trimming_txt=trimming_txt,
                    n_components=n_components,
                    solver=solver,
                    drift_type=drift_type,
                    init_params=init_params,
                    real_bounds=real_bounds,
                    niter=niter,
                    seed=seed,
                    class_code=class_code,
                    shared_memory=shared_memory,
                    nnn=nnn,
                    debug=False,
                    )

        else:
            from time import sleep
            logger.info("dummy started with niter=%d, seed=%s", niter, str(seed))
            for k in range(int(sleep_seconds)):
                if k % 3 == 0:
                    with open(callback_txt, "a") as fh:
                        fh.write(str([k])+"\n")
                sleep(1)
        if shared_memory is not None:
            shared_memory.close()
            logger.info("shared_memory closed")
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(logger, "main_impl failed: ", n=10)
        exit(-1)

if __name__ == '__main__':
    main()



"""
    GroupingMatrix.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from TkMiniTable import TkMiniTable
import tkinter.font as font

SELECTED_BG = 'green'

def num_to_char(i):
    if i is None:
        # occurred in tick formatter
        return ''
    else:
        return hex(i)[2]

def get_font_size(n):
    m = max(0, n-10)
    return 10 - round(m*4/6)

class GroupingMatrix(Tk.Frame):
    def __init__(self, parent, nrows=10, ncols=10, read_only=False):
        Tk.Frame.__init__(self, parent)

        table_frame = Tk.Frame(self)
        table_frame.pack()

        table = TkMiniTable(table_frame,
            columns = ['_'] + ['C%s' % num_to_char(j) for j in range(ncols)],
            default_colwidth=2,
            font = font.Font(self, family="Ariel", size=get_font_size(nrows)),
            cell_widget_class=Tk.Label,
            read_only=read_only,
            on_cell_click=None if read_only else self.on_cell_click,
            on_cell_click_only=True,
            )

        table.heading('_', text='G*' )
        table.column ('_', justify=Tk.CENTER, width=2)
        for j in range(ncols):
            Cn = 'C%s' % num_to_char(j)
            table.heading(Cn, text=Cn)

        table.pack()
        for i in range( 0, nrows ):
            Gn = 'G%s' % num_to_char(i)
            table.insert_row( [Gn] + [''] * ncols )

        self.table = table
        self.default_bg = table.cells_array[1,1].widget.cget('bg')
        self.set_default_grouping(nrows, ncols)

    def set_default_grouping(self, nrows, ncols):
        self.select_matrix = np.zeros((nrows, ncols), dtype=int)
        for i in range(nrows):
            self.select_matrix[i,i] = 1
            cell = self.table.cells_array[i+1, i+1]
            cell.widget.config(bg=SELECTED_BG)

    def on_cell_click(self, cell):
        # print('cell', cell)
        row = cell.row - 1
        col = cell.col - 1
        if row < 0 or col < 0:
            return

        selected = self.select_matrix[row, col]
        bg = self.default_bg if selected else SELECTED_BG
        cell.widget.config(bg=bg)

        self.select_matrix[row, col] = 1 - selected
        selected_col = self.select_matrix[:, col]
        if np.sum(selected_col) > 1:
            for i in range(len(selected_col)):
                if i == row:
                    continue
                cell = self.table.cells_array[i+1, col+1]
                cell.widget.config(bg=self.default_bg)
                selected_col[i] = 0

    def click(self, i, j):
        cell = self.table.cells_array[i+1,j+1]
        self.on_cell_click(cell)

    def reset(self):
        nrows = self.select_matrix.shape[0]
        for i in range(nrows):
            if self.select_matrix[i, i]:
                pass
            else:
                self.click(i, i)

    def get_matrix(self):
        return self.select_matrix



"""
    SettingsDialog.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, Font, FileDialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.OurMessageBox as MessageBox
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.ATSAS.AutoRg import set_exe_array, get_autorg_exe_paths
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, ITEM_DEFAULTS, ALTERNATIVE_WEIGHTS, load_settings_dict, save_settings
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.ListingFrame import ListingFrame
import AutorgKek.Settings   # necessary?
from molass_legacy._MOLASS.Version import get_version_string

label_gb_color_var = 'white smoke'
relief_var = Tk.GROOVE

SHOW_QUALITY_FACTOR_WEIGHTS = False
USE_SPINBOX_FOR_NUM_POINTS  = True
ENABLE_CHART_AXES_OPTIONS   = False
LATER_THAN_V2 = get_version_string().find("_MOLASS 2") >= 0

class SettingsDialog(Dialog):

    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.applied            = False
        self.temp_absorbance    = None
        self.absorbance_has_been_changed    = False
        self.linsting = None

        Dialog.__init__( self, parent, title )

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self )

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=40, pady=10 )

        text_entry_width    = 12
        state_show_sizable_width    = 163
        state_show_sizable_height   = 20

        grid_row = -1

        # ATSAS autorg paths ---------------------------------------------------
        grid_row += 1
        atsas_autorg_paths_label    = Tk.Label( iframe, text= 'ATSAS Version Selection' )
        atsas_autorg_paths_label.grid( row=grid_row, column=0, sticky=Tk.NW ) 

        self.linting_grid_row = grid_row
        self.iframe = iframe
        atsas_exe_paths = get_setting( 'atsas_exe_paths' )

        if len(atsas_exe_paths) == 0:
            atsas_exe_paths = get_autorg_exe_paths()

        self.refresh_atsas_exe_path_listing(atsas_exe_paths)
        grid_row += 1
        label = Tk.Label(iframe, text="Drag one of the above entries to change the reference order.")
        label.grid(row=grid_row, column=1)

        grid_row += 1
        path_update_frame = Tk.Frame(iframe)
        path_update_frame.grid(row=grid_row, column=1)
        label = Tk.Label(path_update_frame, text="Press the button right to update the above paths.")
        label.grid(row=0, column=0)
        button = Tk.Button(path_update_frame, text="Update", command=self.update_atsas_exe_paths)
        button.grid(row=0, column=1)

        # Absorbance-Intensity Mapping  ----------------------------------------
        grid_row += 1
        spacing_frame = Tk.Frame( iframe, height=20 )
        spacing_frame.grid( row=grid_row, column=0 )

        grid_row += 1
        absorbance_intensity_mapping_label = Tk.Label( iframe, text= 'Curves on the Standard Mapping Plane' )
        absorbance_intensity_mapping_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        picking_label_frame = Tk.Frame( iframe )
        picking_label_frame.grid( row=grid_row, column=0, sticky=Tk.E )
        picking_value_frame = Tk.Frame( iframe )
        picking_value_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        picking_row = 0

        # Picking Absorbance Wave Lenghth  -------------------------------------
        absorbance_picking_label    = Tk.Label( picking_label_frame, text= 'UV Absorbance Elution Curve:   ')
        absorbance_picking_label.grid( row=picking_row, column=0, sticky=Tk.E )

        num_points_label    = Tk.Label( picking_value_frame, text= 'averaging')
        num_points_label.grid( row=picking_row, column=0, sticky=Tk.W )

        self.num_points_abs  = Tk.IntVar()
        self.num_points_abs.set( get_setting('num_points_absorbance') )
        if USE_SPINBOX_FOR_NUM_POINTS:
            num_points_abs_entry    = Tk.Spinbox( picking_value_frame, textvariable=self.num_points_abs,  width=3, justify=Tk.CENTER,
                                        from_=1, to=15, increment=2 )
        else:
            num_points_abs_entry    = Tk.Entry( picking_value_frame, textvariable=self.num_points_abs, width=3, justify=Tk.CENTER )
        num_points_abs_entry.grid( row=picking_row, column=1, sticky=Tk.W )
        self.num_points_abs.trace( 'w', self.num_points_abs_tracer )

        around_label = Tk.Label( picking_value_frame, text= 'points around')
        around_label.grid( row=picking_row, column=2, sticky=Tk.W )

        wavelength_label= Tk.Label( picking_value_frame, text= 'picking  wave length λ₁=')
        wavelength_label.grid( row=picking_row, column=3, sticky=Tk.E )
        self.absorbance_picking     = Tk.DoubleVar()
        self.absorbance_picking.set( get_setting('absorbance_picking') )
        absorbance_picking_entry    = Tk.Entry( picking_value_frame, textvariable=self.absorbance_picking, width=8, justify=Tk.CENTER )
        absorbance_picking_entry.grid( row=picking_row, column=4, sticky=Tk.W )
        useage_label = Tk.Label( picking_value_frame, text= '(used in the analysis)' )
        useage_label.grid( row=picking_row, column=5, sticky=Tk.W, padx=5 )

        for i in range(3):
            picking_row += 1
            spacing_label = Tk.Label( picking_label_frame, text=" " )
            spacing_label.grid(row=picking_row, column=0)

        self.zero_absorbance = Tk.DoubleVar()
        self.zero_absorbance.set( get_setting('zero_absorbance') )
        wavelength_label= Tk.Label( picking_value_frame, text= 'baseline wave length λ₂=')
        wavelength_label.grid( row=picking_row, column=3, sticky=Tk.E )
        zero_absorbance_entry    = Tk.Entry( picking_value_frame, textvariable=self.zero_absorbance, width=8, justify=Tk.CENTER )
        zero_absorbance_entry.grid( row=picking_row, column=4, sticky=Tk.W )
        useage_label = Tk.Label( picking_value_frame, text= '(used in the analysis)' )
        useage_label.grid( row=picking_row, column=5, sticky=Tk.W, padx=5  )

        picking_row += 1
        self.zero_absorbance_auto = Tk.IntVar()
        self.zero_absorbance_auto.set(get_setting("zero_absorbance_auto"))
        w = Tk.Checkbutton(picking_value_frame, text="baseline wave length auto correction", variable=self.zero_absorbance_auto)
        w.grid(row=picking_row, column=3, columnspan=3, sticky=Tk.W)

        picking_row += 1
        self.absorbance_picking_sub     = Tk.DoubleVar()
        self.absorbance_picking_sub.set( get_setting('absorbance_picking_sub') )
        wavelength_label= Tk.Label( picking_value_frame, text= 'observe wave length λ₃=')
        wavelength_label.grid( row=picking_row, column=3, sticky=Tk.E )
        absorbance_picking_entry    = Tk.Entry( picking_value_frame, textvariable=self.absorbance_picking_sub, width=8, justify=Tk.CENTER )
        absorbance_picking_entry.grid( row=picking_row, column=4, sticky=Tk.W )
        useage_label = Tk.Label( picking_value_frame, text= '(for observation only)' )
        useage_label.grid( row=picking_row, column=5, sticky=Tk.W, padx=5  )

        # btn_state = Tk.DISABLED if self.parent.serial_data is None else Tk.NORMAL
        if False:
            btn_state = Tk.DISABLED
            show_fig_button = Tk.Button( iframe, text='Figure', command=self.show_absorbance_figure, state=btn_state )
            show_fig_button.grid( row=grid_row, column=1, sticky=Tk.E )

        self.absorbance_picking.trace( 'w', self.absorbance_update_tracer )
        self.zero_absorbance.trace( 'w', self.absorbance_update_tracer )
        self.zero_absorbance_auto.trace( 'w', self.absorbance_update_tracer )

        # Picking Intensity Sacttering Vector  ---------------------------------
        grid_row += 1
        picking_row += 1
        intensity_picking_label   = Tk.Label( picking_label_frame, text= 'X-ray Scattering Elution Curve:  ')
        intensity_picking_label.grid( row=picking_row, column=0, sticky=Tk.E )

        num_points_label    = Tk.Label( picking_value_frame, text= 'averaging')
        num_points_label.grid( row=picking_row, column=0, sticky=Tk.W )

        self.num_points_int  = Tk.IntVar()
        self.num_points_int.set( get_setting('num_points_intensity') )

        if USE_SPINBOX_FOR_NUM_POINTS:
            num_points_int_entry    = Tk.Spinbox( picking_value_frame, textvariable=self.num_points_int, width=3, justify=Tk.CENTER,
                                        from_=1, to=15, increment=2 )
        else:
            num_points_int_entry    = Tk.Entry( picking_value_frame, textvariable=self.num_points_int, width=3, justify=Tk.CENTER )
        num_points_int_entry.grid( row=picking_row, column=1, sticky=Tk.W )
        self.num_points_int.trace('w', self.num_points_int_tracer)

        around_label    = Tk.Label( picking_value_frame, text= 'points around')
        around_label.grid( row=picking_row, column=2, sticky=Tk.W )

        qlabel= Tk.Label( picking_value_frame, text= 'scattering vector q=')
        qlabel.grid( row=picking_row, column=3, sticky=Tk.E )

        self.intensity_picking    = Tk.DoubleVar()
        self.intensity_picking.set( get_setting('intensity_picking') )
        intensity_picking_entry   = Tk.Entry( picking_value_frame, textvariable=self.intensity_picking, width=8, justify=Tk.CENTER )
        intensity_picking_entry.grid( row=picking_row, column=4, sticky=Tk.W )

        # Default Angle Range Start -----------------------------------------
        grid_row += 1
        spacing_frame = Tk.Frame( iframe, height=20 )
        spacing_frame.grid( row=grid_row, column=0 )

        grid_row += 1
        angular_range_start_label = Tk.Label(iframe, text= 'Guinier Limit in Small Angle Data Trimming')
        angular_range_start_label.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W )

        grid_row += 1
        self.cut_before_guinier = Tk.IntVar()
        self.cut_before_guinier.set( get_setting('cut_before_guinier') )
        texts = [   '(1) Minimum of the measured Q range',
                    '(2) Minimum of the Guinier Q range',
                    '(3) Extended Limit beyond the Guinier Q range',
                ]
        for k, text in enumerate(texts):
            grid_row += 1
            rb_frame = Tk.Frame( iframe )
            rb_frame.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=30 )
            rb = Tk.Radiobutton( rb_frame, text=text, variable=self.cut_before_guinier, value=k )
            rb.grid( row=0, column=0, sticky=Tk.W + Tk.N )
            if k == 2:
                limit_prob_frame = Tk.Frame(rb_frame)
                limit_prob_frame.grid( row=0, column=1, padx=30)
                limit_prob_label = Tk.Label(limit_prob_frame, text="Acceptable Rg-consistency: ")
                limit_prob_label.grid( row=0, column=0 )
                self.acceptable_rg_consist = Tk.DoubleVar()
                self.acceptable_rg_consist.set( get_setting('acceptable_rg_consist') )
                limit_prob_entry = Tk.Spinbox(limit_prob_frame, textvariable=self.acceptable_rg_consist, width=6, justify=Tk.CENTER,
                                        from_=0, to=1, increment=0.1)
                limit_prob_entry.grid( row=0, column=1, sticky=Tk.W )
                guide_label_min = Tk.Label(limit_prob_frame, text="Setting 0 here implies the above option (1)")
                guide_label_min.grid( row=1, column=1, padx=5 )
                guide_label_max = Tk.Label(limit_prob_frame, text="Setting 1 here implies the above option (2)")
                guide_label_max.grid( row=2, column=1, padx=5 )

        # Keeping Temporary Folders
        grid_row += 1
        spacing_frame = Tk.Frame(iframe, height=20)
        spacing_frame.grid(row=grid_row, column=0)

        grid_row += 1
        saving_temporary_folders_label = Tk.Label(iframe, text="Keeping Temporary Folders")
        saving_temporary_folders_label.grid(row=grid_row, column=0, columnspan=2, sticky=Tk.W)

        grid_row += 1
        cb_frame = Tk.Frame( iframe )
        cb_frame.grid(row=grid_row, column=0, columnspan=2, sticky=Tk.W, padx=30)

        self.keep_tempfolder_averaged = Tk.IntVar()
        self.keep_tempfolder_averaged.set(get_setting("keep_tempfolder_averaged"))

        w = Tk.Checkbutton(cb_frame, text='keep "averaged" folder used to run ALMERGE', variable=self.keep_tempfolder_averaged)
        w.grid(row=0, column=0, sticky=Tk.W)

        # Default Font of Result Excel Books
        grid_row += 1
        spacing_frame = Tk.Frame(iframe, height=20)
        spacing_frame.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Default Font of Result Excel Books")
        label.grid(row=grid_row, column=0, columnspan=2, sticky=Tk.W)

        grid_row += 1
        self.report_default_font = Tk.StringVar()
        self.report_default_font.set(get_setting("report_default_font"))
        self.font_box = ttk.Combobox(master=iframe, values=["Arial", "Calibri", "Times New Roman", "ＭＳ Ｐゴシック"],
                                        textvariable=self.report_default_font, width=16)
        self.font_box.grid(row=grid_row, column=0, sticky=Tk.W, padx=30)

        # Quality Factor Weights
        factor_names = []
        if SHOW_QUALITY_FACTOR_WEIGHTS:
            # Quality Factor Weighting ---------------------------------------------
            grid_row += 1
            spacing_frame = Tk.Frame( iframe, height=20 )
            spacing_frame.grid( row=grid_row, column=0 )

            grid_row += 1
            quality_factor_weights_label = Tk.Label( iframe, text= 'Quality Factor Weighting' )
            quality_factor_weights_label.grid( row=grid_row, column=0, sticky=Tk.W )

            factor_names += [ 'Basic Qualilty', 'Positive Score', 'Rg Stdev Score', 'Left q*Rg Score', 'End Consistency' ]
            # quality_weighting = Settings.get_setting( 'quality_weighting' )
            quality_weighting = get_setting( 'quality_weighting' )
            self.quality_factors = []
            for i, name in enumerate( factor_names ):
                grid_row += 1
                factor_label = Tk.Label( iframe, text=name )
                factor_label.grid( row=grid_row, column=0, sticky=Tk.E, padx=10 )
                quality_factor = Tk.DoubleVar()
                self.quality_factors.append( quality_factor )
                quality_factor.set( quality_weighting[i] )
                if i == 2:
                    quality_factor_entry = Tk.Frame( iframe )
                    entry_widget = Tk.Entry( quality_factor_entry, textvariable=quality_factor, width=8, justify=Tk.CENTER, state=Tk.DISABLED )
                    entry_widget.grid( row=0, column=0, sticky=Tk.W )
                    default_weights_button = Tk.Button( quality_factor_entry, text="Default Weights", command=self.default_weights, state=Tk.DISABLED )
                    default_weights_button.grid( row=0, column=1, sticky=Tk.W, padx=80 )
                else:
                    quality_factor_entry = Tk.Entry( iframe, textvariable=quality_factor, width=8, justify=Tk.CENTER, state=Tk.DISABLED )
                quality_factor_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        # Chart axis direction for descending side --------------------------
        if ENABLE_CHART_AXES_OPTIONS:
            grid_row += 1
            spacing_frame = Tk.Frame( iframe, height=20 )
            spacing_frame.grid( row=grid_row, column=0 )

            grid_row += 1
            axis_direction_desc_label = Tk.Label( iframe, text= 'Chart axes direction for descending sides' )
            axis_direction_desc_label.grid( row=grid_row, column=0, sticky=Tk.W )

            self.axis_direction_desc = Tk.IntVar()
            self.axis_direction_desc.set( get_setting( 'axis_direction_desc' ) )
            grid_row += 1

            tail_text = ' order of concentration'
            for i, t in enumerate( [ 'Descending'+ tail_text, 'Ascending' + tail_text ] ):
                b = Tk.Radiobutton( iframe, text=t,
                            variable=self.axis_direction_desc, value=i )
                b.grid( row=grid_row+i, column=0, sticky=Tk.W,  padx=30  )

        # Reset to defaults button
        grid_row += len(factor_names) + 4
        spacing_frame = Tk.Frame( iframe, height=20 )
        spacing_frame.grid( row=grid_row, column=0 )

        grid_row += 1
        reset_button = Tk.Button( iframe, text="Reset to Defaults", command=self.reset_to_defaults )
        reset_button.grid( row=grid_row, column=1, sticky=Tk.E )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

        # self.resizable(width=False, height=False)

    def absorbance_update_tracer( self, *args ):
        print( 'absorbance updated' )
        self.absorbance_has_been_changed = True

    def spinbox_tracer(self, var):
        try:
            n = var.get()
        except:
            return

        if n % 2 == 0:
            if n > 5:
                n_ = n - 1
            else:
                n_ = n + 1
            var.set(n_)

    def num_points_abs_tracer( self, *args ):
        self.spinbox_tracer(self.num_points_abs)

    def num_points_int_tracer( self, *args ):
        self.spinbox_tracer(self.num_points_int)

    def reset_to_defaults( self ):
        self.num_points_abs.set(            ITEM_DEFAULTS['num_points_absorbance'] )
        self.absorbance_picking.set(        ITEM_DEFAULTS['absorbance_picking'] )
        self.zero_absorbance.set(           ITEM_DEFAULTS['zero_absorbance'] )
        self.zero_absorbance_auto.set(      ITEM_DEFAULTS['zero_absorbance_auto'] )
        self.absorbance_picking_sub.set(    ITEM_DEFAULTS['absorbance_picking_sub'] )
        self.num_points_int.set(            ITEM_DEFAULTS['num_points_intensity'] )
        self.intensity_picking.set(         ITEM_DEFAULTS['intensity_picking'] )
        if ENABLE_CHART_AXES_OPTIONS:
            self.axis_direction_desc.set(   ITEM_DEFAULTS['axis_direction_desc'] )
        self.cut_before_guinier.set(        ITEM_DEFAULTS['cut_before_guinier'] )
        self.acceptable_rg_consist.set(     ITEM_DEFAULTS['acceptable_rg_consist'] )
        self.keep_tempfolder_averaged.set(  ITEM_DEFAULTS['keep_tempfolder_averaged'] )
        self.report_default_font.set(       ITEM_DEFAULTS['report_default_font'] )

    def buttonbox(self):
        '''
        override standard buttonbox.
        add "Default" button
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        """
        w = Tk.Button(box, text="Default", width=10, command=self.reset_to_default)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        """

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack( pady=10 )

    def refresh_atsas_exe_path_listing(self, atsas_exe_paths):
        if self.linsting is not None:
            self.linsting.destroy()

        grid_row = self.linting_grid_row
        self.linsting = ListingFrame(self.iframe, numbering=True)
        self.linsting.grid(row=grid_row, column=1)

        self.path_var_array = []
        num_path_entries = max( 2, len( atsas_exe_paths ) )
        for i in range(num_path_entries):
            if i < len( atsas_exe_paths ):
                path = atsas_exe_paths[i]
            else:
                path = ''
            var = Tk.StringVar()
            var.set(path.replace("\\autorg.exe", ""))
            self.path_var_array.append( var )
            folder_entry = FolderEntry(self.linsting, textvariable=var, slimbutton=True, width=80)
            self.linsting.insert(var, folder_entry)

        self.linsting.update_bind_geometry()

    def update_atsas_exe_paths(self):
        atsas_exe_paths = get_autorg_exe_paths()
        self.refresh_atsas_exe_path_listing(atsas_exe_paths)

    def show_absorbance_figure( self ):
        from molass_legacy.UV.AbsorbancePlot import show_absorbance_figure_util
        show_absorbance_figure_util( self )

    def apply( self ):  # overrides parent class method
        print( "ok. apply" )

        atsas_exe_paths = []
        for var in self.linsting.get_variables():
            path = var.get()
            atsas_exe_paths.append( path + r'\autorg.exe' )

        ok_ = True

        try:
            absorbance_picking  = self.absorbance_picking.get()
            zero_absorbance     = self.zero_absorbance.get()
            zero_absorbance_auto    = self.zero_absorbance_auto.get()
            absorbance_picking_sub  = self.absorbance_picking_sub.get()
            intensity_picking   = self.intensity_picking.get()
            num_points_abs      = self.num_points_abs.get()
            num_points_int      = self.num_points_int.get()
            if SHOW_QUALITY_FACTOR_WEIGHTS:
                weights = []
                for factor_var in self.quality_factors:
                    weights.append( factor_var.get() )

        except:
            ok_ = False
            error_msg_param = [ 'Value Format Error', 'There exits an ill-formatted value.' ]

        if ok_ and num_points_abs > 0 and num_points_int > 0:
            pass
        else:
            ok_ = False
            error_msg_param = [ 'Value Range Error', 'Number of points must be positive.' ]

        if ok_ and absorbance_picking > 0 and intensity_picking > 0:
            pass
        else:
            ok_ = False
            error_msg_param = [ 'Value Range Error', 'Picking parameter must be positive.' ]

        if SHOW_QUALITY_FACTOR_WEIGHTS:
            if ok_:
                sum_w = np.sum( weights )
                if abs( sum_w - 1 ) < 1e-10:
                    pass
                else:
                    ok_ = False
                    error_msg_param = [ 'Value Range Error', 'Sum of the quality factors must be 1.' ]

        if not ok_:
            MessageBox.showerror(
                error_msg_param[0],
                error_msg_param[1],
                parent=self,
                )
            self.applied    = None
            # setting this to None causes retry in the parent
            return

        set_exe_array(paths=atsas_exe_paths)
        set_setting( 'absorbance_picking',      absorbance_picking )
        set_setting( 'zero_absorbance',         zero_absorbance )
        set_setting( 'zero_absorbance_auto',    zero_absorbance_auto )
        set_setting( 'absorbance_picking_sub',  absorbance_picking_sub )
        set_setting( 'intensity_picking',       intensity_picking )
        set_setting( 'num_points_absorbance',   num_points_abs )
        set_setting( 'num_points_intensity',    num_points_int )
        set_setting( 'cut_before_guinier',      self.cut_before_guinier.get() )
        set_setting( 'acceptable_rg_consist',   self.acceptable_rg_consist.get() )
        set_setting( 'keep_tempfolder_averaged',   self.keep_tempfolder_averaged.get() )
        set_setting( 'report_default_font',     self.report_default_font.get() )

        if SHOW_QUALITY_FACTOR_WEIGHTS:
            Settings.set_setting( 'quality_weighting',  weights )

        if ENABLE_CHART_AXES_OPTIONS:
            set_setting( 'axis_direction_desc',     self.axis_direction_desc.get() )

        if self.absorbance_has_been_changed and self.temp_absorbance is not None:
            self.parent.serial_data.absorbance = self.temp_absorbance

        save_settings()
        self.applied    = True

    def select_autorg_path( self, i ):
        dir_ = 'C:/'
        f = FileDialog.askopenfilename(
                        initialdir=dir_,
                        filetypes = [
                            ( 'Executalbe file', '*.exe' ),
                            ],
                        parent=self,
          )
        if not f:
            return

        self.path_var_array[i].set( f )

class RestoreSettingDialog( Dialog ):
    def __init__( self, parent, title ):
        self.grab = 'local'     # used in grab_set
        self.parent             = parent
        self.applied            = False

        Dialog.__init__( self, parent, title )

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self )

        grid_row = 0

        result_folder_label  = Tk.Label( body_frame, text= 'Result Folder: ' )
        result_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.result_folder = Tk.StringVar()
        result_folder = self.parent.an_folder.get() + '/' + self.parent.analysis_name.get()
        self.result_folder.set( result_folder )
        self.result_folder_entry = FolderEntry( body_frame, textvariable=self.result_folder, width=70,
                                            on_entry_cb=self.on_entry_in_folder )
        self.result_folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

    def on_entry_in_folder( self ):
        pass

    def validate( self ):
        pickle_file = self.result_folder.get() + '/.save/serial_settings.dump'
        if os.path.exists( pickle_file ):
            self.result_folder_entry.config( fg='black' )
            yn = MessageBox.askyesno( 'Comfirmation',
                            'This action restores settings from the previous results,'
                            '\nand your current input settings will be lost.'
                            '\nAre you sure to proceed?',
                            parent=self
                            )
            if yn:
                self.restore_setting( pickle_file )
                return 1
            else:
                return 0
        else:
            self.result_folder_entry.config( fg='red' )
            MessageBox.showerror( 'Input Error',
                            pickle_file + '\ndoes not exist.' + '\nEnter a folder which really contains results.',
                            parent=self )
            return 0

    def apply( self ):  # overrides parent class method
        print( "ok. apply" )
        self.applied = True

    def restore_setting( self, pickle_file ):
        dict_ = load_settings_dict( pickle_file )
        parent = self.parent
        parent.in_folder.set( dict_.get( 'in_folder' ) )
        parent.uv_folder.set( dict_.get( 'uv_folder' ) )
        parent.uv_file.set( dict_.get( 'uv_file' ) )
        parent.an_folder.set( dict_.get( 'an_folder' ) )
        # TODO: what to restore
        # move this method to parent



# coding: utf-8
"""
    Rgg.Callbacks.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from pomegranate.callbacks import Callback

class VisualizationtCallback(Callback):
    def __init__(self):
        self.model = None
        self.params = None

    def on_epoch_end(self, logs):
        epoch = logs['epoch']
        print("{} ----------- on_epoch_end".format(epoch))



# coding: utf-8
"""
    KillableThread.py

    borrowed from
    http://tomerfiliba.com/recipes/Thread2/

    Copyright (c) 2019-2020, Masatsuyo Takahashi, KEK-PF
"""

import threading
import inspect
import ctypes

def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        # """if it returns a number greater than one, you're in trouble, 
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemError("PyThreadState_SetAsyncExc failed")


class Thread(threading.Thread):
    def _get_my_tid(self):
        """determines this (self's) thread id"""
        if not self.is_alive():
            raise threading.ThreadError("the thread is not active")
        
        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id
        
        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid
        
        raise AssertionError("could not determine the thread's id")
    
    def raise_exc(self, exctype):
        """raises the given exception type in the context of this thread"""
        _async_raise(self._get_my_tid(), exctype)
    
    def terminate(self):
        """raises SystemExit in the context of the given thread, which should 
        cause the thread to exit silently (unless caught)"""
        self.raise_exc(SystemExit)



"""
    MappingOptimizer.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy import optimize
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt

MAJOR_PEAK_THRESHOLD_RATIO  = 0.3
USE_MODEL_PEAKS = True
USE_SLICE = True

def draw_zoomed_peaks_for_proof(a_curve, x_curve):
    from molass_legacy.Elution.CurveUtils import simple_plot
    from DataUtils import get_in_folder
    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle("Proof Plot of the Poor Mapping on " + get_in_folder(), fontsize=20)
        ax1.set_title("UV Elution Curve", fontsize=16)
        ax2.set_title("XR Elution Curve", fontsize=16)

        for ax, ecurve, color in [(ax1, a_curve, "blue"), (ax2, x_curve, "orange")]:
            simple_plot(ax, ecurve, color=color, legend=False)

            # inset axes....
            x = ecurve.x
            y = ecurve.y
            j = np.argmax(y)
            px = x[j]
            py = y[j]
            width = ecurve.get_peak_region_width()
            dx = width/8
            dy = py*0.05
            axins = ax.inset_axes(
                [0.58, 0.3, 0.4, 0.4],
                xlim=(px-dx, px+dx), ylim=(py-2*dy, py+dy), xticklabels=[], yticklabels=[])
            axins.grid(False)
            # axins.set_facecolor('gray')
            for name in ['top', 'bottom', 'left', 'right']:
                axins.spines[name].set_color('black')

            for epeak in ecurve.get_emg_peaks():
                y_ = epeak.get_model_y(x)
                axins.plot(x, y_, color="cyan", lw=3)
                ymin, ymax = axins.get_ylim()
                cx = epeak.opt_params[1]
                axins.plot([cx, cx], [ymin, ymax], lw=3, color="yellow", label=r"EGH $\mu$")
                axins.legend(loc="upper left")

            simple_plot(axins, ecurve, color=color, legend=False)

            ax.indicate_inset_zoom(axins, edgecolor="black")
            ax.legend()

        fig.tight_layout()
        plt.show()

def make_slice_for_optimization(ecurve, debug=True):
    f, t = ecurve.get_peak_region_sigma()
    i, j = [bisect_right(ecurve.x, p) for p in [f, t]]
    return slice(i,j+1)

def get_reliable_peak_top_x(ecurve):
    # consider doing this earlier
    ret_x = []
    for epeak in ecurve.get_emg_peaks():
        ret_x.append(epeak.opt_params[1])
    return ret_x

def optimize_mapping_impl(self, debug=False):

    self.determine_mapping_ranges()

    self.inv_mapped_boundaries = []
    tgt_y = self.x_curve.y
    xray_x = self.x_curve.x
    uv_x = self.a_curve.x

    assert len(self.a_curve.peak_top_x) == len(self.x_curve.peak_top_x)
    threshold = self.x_curve.max_y * MAJOR_PEAK_THRESHOLD_RATIO

    if debug:
        draw_zoomed_peaks_for_proof(self.a_curve, self.x_curve)

    if USE_MODEL_PEAKS:
        xr_candidates = get_reliable_peak_top_x(self.x_curve)
        uv_candidates = get_reliable_peak_top_x(self.a_curve)
    else:
        xr_candidates = self.x_curve.peak_top_x
        uv_candidates = self.a_curve.peak_top_x

    xr_top_x = []
    xr_top_y = []
    uv_top_x = []
    for k, px in enumerate(xr_candidates):
        py = self.x_curve.spline(px)
        if py > threshold:
            xr_top_x.append(px)
            xr_top_y.append(py)
            uv_top_x.append(uv_candidates[k])
    xr_top_x = np.array(xr_top_x)
    xr_top_y = np.array(xr_top_y)
    peak_weights = xr_top_y/np.sum(xr_top_y)
    uv_top_x = np.array(uv_top_x)
    num_peaks = len(xr_top_x)

    if USE_SLICE:
        slice_for_opt = make_slice_for_optimization(self.x_curve)
        if debug:
            from matplotlib.patches import Rectangle
            from molass_legacy.Elution.CurveUtils import simple_plot
            from DataUtils import get_in_folder
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("Range for Mapping Optmization on " + get_in_folder())
                simple_plot(ax, self.x_curve, legend=False)
                ymin, ymax = ax.get_ylim()
                f, t = slice_for_opt.start, slice_for_opt.stop
                p_ = Rectangle(
                        (f, ymin),      # (x,y)
                        t - f,          # width
                        ymax - ymin,    # height
                        facecolor='cyan',
                        alpha=0.2,
                        label="Optimization Range",
                    )
                ax.add_patch(p_)
                ax.legend()
                fig.tight_layout()
                plt.show()
    else:
        slice_for_opt = slice(None, None)

    tgt_y_ = tgt_y[slice_for_opt]

    def objective(p, debug_info=None):
        A, B = p[0:2]
        mapped_y, _ = self.make_whole_mapped_vector(A, B, p[2:], make_inv_mapped_boundaries=True, with_original_scale=True)
        msd = np.average((mapped_y[slice_for_opt] - tgt_y_)**2)
        if num_peaks == 1:
            psd = 1
        else:
            psd = np.average((peak_weights*(A*xr_top_x + B - uv_top_x))**2)

        if debug_info is not None:
            title, = debug_info
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(title)
                ax.plot(xray_x, tgt_y, label="tgt_y")
                ax.plot(xray_x, mapped_y, label="mapped_y")
                ymin, ymax = ax.get_ylim()
                if USE_SLICE:
                    from matplotlib.patches import Rectangle
                    f, t = slice_for_opt.start, slice_for_opt.stop
                    p_ = Rectangle(
                            (f, ymin),      # (x,y)
                            t - f,          # width
                            ymax - ymin,    # height
                            facecolor   = 'cyan',
                            alpha       = 0.2,
                        )
                    ax.add_patch(p_)
                ax.set_ylim(ymin, ymax)
                for xrx, uvx in zip(xr_top_x, (uv_top_x - B)/A):
                    ax.plot([xrx, xrx], [ymin, ymax], color="orange", label="XR peak top")
                    ax.plot([uvx, uvx], [ymin, ymax], color="blue", label="UV peak top")
                ax.legend()
                fig.tight_layout()
                plt.show()

        return msd*psd

    if debug:
        self.logger.info("feature_mapped=%s, manual_sync=%s", str(self.feature_mapped), str(self.manual_sync))

    if self.feature_mapped or self.manual_sync:
        if self.manual_sync:
            from molass_legacy._MOLASS.SerialSettings import get_setting
            A_ = get_setting('manual_time_scale')
            B_ = get_setting('manual_time_shift')
            A_init = 1/A_
            B_init = -B_/A_
        else:
            A_init = self.A_init
            B_init = self.B_init

        init_params = (A_init, B_init, *self.init_scales)
        delta = 1e-10   # trick to make the code simple for these options
        ab_bounds = ((A_init-delta, A_init+delta), (B_init-delta, B_init+delta))
    else:
        b_delta = uv_x[-1]*0.1
        init_params = (self.A_init, self.B_init, *self.init_scales)
        ab_bounds = ((self.A_init*0.9, self.A_init*1.1), (self.B_init-b_delta, self.B_init+b_delta))

    scale_bounds = [(0.1*s, 10*s) for s in self.init_scales]
    bounds = (*ab_bounds, *scale_bounds)

    if debug:
        objective(init_params, debug_info=("before minimize",))

    ret = optimize.minimize(objective, init_params, bounds=bounds)

    if debug:
        objective(ret.x, debug_info=("after minimize",))

    A, B = ret.x[0:2]

    if self.manual_sync:
        new_A_ = 1/A
        new_B_ = -B/A
        self.logger.info("manual params unchanged proof: (%g, %g) ⇔ (%g, %g)", A_, B_, new_A_, new_B_)

    self.mapped_vector, self.mapped_vector_orig = self.make_whole_mapped_vector(A, B, ret.x[2:], make_inv_mapped_boundaries=True, with_original_scale=True )
    self.map_params = A, B
    self.scale_params = ret.x[2:]

    # this is temporary for backward compatibility and should be removed in the near future
    temp_results = []
    A, B = self.map_params
    for S in self.scale_params:
        temp_results.append((A, B, S))
    self.opt_results = temp_results

    self.uv_ranges = []
    for i, range_ in enumerate( self.mapping_ranges ):
        self.uv_ranges.append( [ A*x + B for x in range_ ] )

    assert len(self.mapped_vector) == len(self.x_x)

    self.mapped_spline  = UnivariateSpline( self.x_x, self.mapped_vector, s=0, ext=3 )

    if False:
        from CallStack import CallStack
        cs = CallStack()
        print(cs)
        print( 'len(self.x_x)=', len(self.x_x) )
        print( 'len(self.mapped_vector)=', len(self.mapped_vector) )
        ax = plt.gca()
        ax.cla()
        ax.plot( self.x_x, self.x_curve.y, color='orange', label='x_curve.y' )
        ax.plot( self.mapped_vector, color='blue', label='mapped_vector' )
        ax.plot( self.x_x, self.mapped_spline(self.x_x), color='green', label='mapped_spline' )
        ax.legend()
        plt.tight_layout()
        plt.show()

    assert len(self.x_x) == len(self.mapped_vector)
